# -*- coding: utf-8 -*-
#
# This document is free and open-source software, subject to the OSI-approved
# BSD license below.
#
# Copyright (c) 2011 - 2013 Alexis Petrounias <www.petrounias.org>,
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the author nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

""" Unit tests for the Django CTE Trees test application.
"""

from datetime import date, timedelta
from uuid import UUID

from django.core.exceptions import FieldError, ImproperlyConfigured
from django.db.models import Avg, F
from django.test import TestCase

from cte_forest_test.models import (
    AggregationNode, BadParameter_delete_Node, BadParameter_parent_1_Node,
    BadParameter_parent_2_Node, BadParameter_parent_3_Node,
    BadParameter_parent_4_Node, BadParameter_traversal_Node, BFSOrderedNode,
    CustomPrimaryKeyNode, DBTypeNode, DBTypePrimaryKeyNode, DFSOrderedNode,
    ExoticTypeNode, NoneDeleteNode, NoneTraversalNode, OrderedNamedNode,
    SimpleNamedNode, SimpleNamedNodeUser, SimpleNode, SimpleNodeUser,
    TypeCoercionNode, TypeCombinationNode, ValueNamedNode,
)


class SimpleNodeTest(TestCase):

    def test_node_creation(self):

        node = SimpleNode.objects.create()

        self.assertEqual(SimpleNode.objects.get().id, node.id)
        self.assertEqual(len(SimpleNode.objects.all()), 1)
        self.assertEqual(len(SimpleNode.objects.filter()), 1)
        self.assertEqual(SimpleNode.objects.count(), 1)

    def test_node_creation_save(self):

        node = SimpleNode()
        node.save()

        self.assertEqual(SimpleNode.objects.get().id, node.id)
        self.assertEqual(len(SimpleNode.objects.all()), 1)
        self.assertEqual(len(SimpleNode.objects.filter()), 1)
        self.assertEqual(SimpleNode.objects.count(), 1)

    def test_multiple_node_creation(self):

        SimpleNode.objects.create()
        SimpleNode.objects.create()

        self.assertEqual(len(SimpleNode.objects.filter()), 2)
        self.assertEqual(SimpleNode.objects.count(), 2)

    def test_node_save(self):

        node = SimpleNode.objects.create()
        node.save()

        self.assertEqual(node.id, SimpleNode.objects.get().id)
        self.assertEqual(len(SimpleNode.objects.all()), 1)
        self.assertEqual(len(SimpleNode.objects.filter()), 1)
        self.assertEqual(SimpleNode.objects.count(), 1)

    def test_node_delete(self):

        node = SimpleNode.objects.create()

        node.delete()

        self.assertEqual(len(SimpleNode.objects.filter()), 0)
        self.assertEqual(SimpleNode.objects.count(), 0)

    def test_node_delete_query(self):

        SimpleNode.objects.create()

        SimpleNode.objects.filter().delete()

        self.assertEqual(len(SimpleNode.objects.filter()), 0)
        self.assertEqual(SimpleNode.objects.count(), 0)

    def test_tree_structure(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node = SimpleNode.objects.create(parent=middle_node)

        fresh_root_node = SimpleNode.objects.get(id=root_node.id)
        fresh_middle_node = SimpleNode.objects.get(id=middle_node.id)
        fresh_bottom_node = SimpleNode.objects.get(id=bottom_node.id)

        self.assertEqual(fresh_root_node.depth, 1)
        self.assertEqual(fresh_root_node.path, [root_node.id])
        self.assertEqual(fresh_root_node.ordering, [root_node.id])

        self.assertEqual(fresh_middle_node.depth, 2)
        self.assertEqual(fresh_middle_node.path,
                         [root_node.id, middle_node.id])
        self.assertEqual(fresh_middle_node.ordering,
                         [root_node.id, middle_node.id])

        self.assertEqual(fresh_bottom_node.depth, 3)
        self.assertEqual(fresh_bottom_node.path,
                         [root_node.id, middle_node.id, bottom_node.id])
        self.assertEqual(fresh_bottom_node.ordering,
                         [root_node.id, middle_node.id, bottom_node.id])

    def test_node_roots(self):

        root_node_1 = SimpleNode.objects.create()
        root_node_2 = SimpleNode.objects.create()
        SimpleNode.objects.create(parent=root_node_1)
        SimpleNode.objects.create(parent=root_node_2)

        self.assertEqual(len(SimpleNode.objects.roots()), 2)
        self.assertEqual(SimpleNode.objects.roots().count(), 2)

        root_node_pks = [root_node_1.pk, root_node_2.pk]
        for node in SimpleNode.objects.roots():
            self.assertTrue(node.pk in root_node_pks)

    def test_tree_structure_leaves(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node_1 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_2 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_3 = SimpleNode.objects.create(parent=middle_node)

        self.assertEqual([bottom_node_1.id, bottom_node_2.id, bottom_node_3.id],
                         [node.id for node in SimpleNode.objects.leaves()])

    def test_tree_structure_branches(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        SimpleNode.objects.create(parent=middle_node)
        SimpleNode.objects.create(parent=middle_node)
        SimpleNode.objects.create(parent=middle_node)

        self.assertEqual(
            {root_node.id, middle_node.id},
            {node.id for node in SimpleNode.objects.branches()})

    def test_tree_structure_parent(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node = SimpleNode.objects.create(parent=middle_node)

        self.assertEqual(root_node.parent, None)
        self.assertEqual(middle_node.parent.id, root_node.id)
        self.assertEqual(bottom_node.parent.id, middle_node.id)
        self.assertEqual(bottom_node.parent.parent.id, root_node.id)

    def test_tree_structure_children(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node_1 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_2 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_3 = SimpleNode.objects.create(parent=middle_node)

        self.assertEqual(root_node.children.all()[0].id, middle_node.id)

        self.assertEqual([node.pk for node in middle_node.children.all()],
                         [bottom_node_1.pk, bottom_node_2.pk, bottom_node_3.pk])

        self.assertEqual(len(bottom_node_1.children.all()), 0)

    def test_tree_structure_siblings(self):

        root_node = SimpleNode.objects.create()
        middle_node_1 = SimpleNode.objects.create(parent=root_node)
        middle_node_2 = SimpleNode.objects.create(parent=root_node)
        bottom_node_1 = SimpleNode.objects.create(parent=middle_node_1)
        bottom_node_2_1 = SimpleNode.objects.create(parent=middle_node_2)
        bottom_node_2_2 = SimpleNode.objects.create(parent=middle_node_2)

        self.assertEqual(len(root_node.siblings()), 0)
        self.assertEqual(root_node.siblings().count(), 0)

        self.assertEqual(middle_node_1.siblings()[0].id, middle_node_2.id)
        self.assertEqual(middle_node_2.siblings()[0].id, middle_node_1.id)

        self.assertEqual(len(bottom_node_1.siblings()), 0)
        self.assertEqual(bottom_node_1.siblings().count(), 0)

        self.assertEqual(bottom_node_2_1.siblings()[0].id, bottom_node_2_2.id)
        self.assertEqual(bottom_node_2_2.siblings()[0].id, bottom_node_2_1.id)

    def test_tree_structure_descendants(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node_1 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_2 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_3 = SimpleNode.objects.create(parent=middle_node)

        self.assertEqual([node.pk for node in root_node.descendants()],
                         [middle_node.pk, bottom_node_1.pk, bottom_node_2.pk,
                          bottom_node_3.pk])

        self.assertEqual([node.pk for node in middle_node.descendants()],
                         [bottom_node_1.pk, bottom_node_2.pk, bottom_node_3.pk])

        self.assertEqual(len(bottom_node_1.descendants()), 0)

    def test_tree_structure_ancestors(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node = SimpleNode.objects.create(parent=middle_node)
        fresh_root_node = SimpleNode.objects.get(id=root_node.id)
        fresh_middle_node = SimpleNode.objects.get(id=middle_node.id)
        fresh_bottom_node = SimpleNode.objects.get(id=bottom_node.id)

        self.assertEqual(len(fresh_root_node.ancestors()), 0)

        self.assertEqual(fresh_root_node.ancestors().count(), 0)

        self.assertEqual(fresh_middle_node.ancestors()[0].id, root_node.id)

        self.assertEqual([node.pk for node in fresh_bottom_node.ancestors()],
                         [root_node.id, middle_node.id])

    def test_tree_structure_root(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node = SimpleNode.objects.create(parent=middle_node)

        fresh_root_node = SimpleNode.objects.get(id=root_node.id)
        fresh_middle_node = SimpleNode.objects.get(id=middle_node.id)
        fresh_bottom_node = SimpleNode.objects.get(id=bottom_node.id)

        self.assertEqual(fresh_root_node.root().id, root_node.id)
        self.assertEqual(fresh_middle_node.root().id, root_node.id)
        self.assertEqual(fresh_bottom_node.root().id, root_node.id)

    def test_tree_structure_intersection(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        SimpleNode.objects.create(parent=middle_node)
        bottom_node_2 = SimpleNode.objects.create(parent=middle_node)
        SimpleNode.objects.create(parent=middle_node)

        # We need the path information to get ancestors.
        fresh_bottom_node_2 = SimpleNode.objects.get(id=bottom_node_2.id)
        q = fresh_bottom_node_2.ancestors() & root_node.descendants()
        self.assertEqual(middle_node.id, q[0].id)

    def test_tree_structure_is_child_of(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node_1 = SimpleNode.objects.create(parent=middle_node)
        SimpleNode.objects.create(parent=middle_node)
        SimpleNode.objects.create(parent=middle_node)

        self.assertTrue(middle_node.is_child_of(root_node))
        self.assertFalse(bottom_node_1.is_child_of(root_node))

    def test_tree_structure_is_parent_of(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node_1 = SimpleNode.objects.create(parent=middle_node)
        SimpleNode.objects.create(parent=middle_node)
        SimpleNode.objects.create(parent=middle_node)

        self.assertTrue(root_node.is_parent_of(middle_node))
        self.assertFalse(root_node.is_parent_of(bottom_node_1))

    def test_tree_structure_is_sibling_of(self):

        root_node = SimpleNode.objects.create()
        another_root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node_1 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_2 = SimpleNode.objects.create(parent=middle_node)
        SimpleNode.objects.create(parent=middle_node)

        self.assertTrue(bottom_node_1.is_sibling_of(bottom_node_2))
        self.assertTrue(bottom_node_2.is_sibling_of(bottom_node_1))

        # Ensure edge case when parents are None.
        self.assertTrue(root_node.is_sibling_of(another_root_node))

        self.assertFalse(bottom_node_1.is_sibling_of(middle_node))
        self.assertFalse(bottom_node_1.is_sibling_of(root_node))

        # Ensure edge case when compared to self.
        self.assertFalse(bottom_node_1.is_sibling_of(bottom_node_1))
        self.assertFalse(middle_node.is_sibling_of(middle_node))

    def test_tree_structure_is_descendant_of(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node = SimpleNode.objects.create(parent=middle_node)

        fresh_root_node = SimpleNode.objects.get(id=root_node.id)
        fresh_middle_node = SimpleNode.objects.get(id=middle_node.id)
        fresh_bottom_node = SimpleNode.objects.get(id=bottom_node.id)

        self.assertTrue(fresh_middle_node.is_descendant_of(fresh_root_node))
        self.assertFalse(fresh_root_node.is_descendant_of(fresh_middle_node))

        self.assertTrue(fresh_bottom_node.is_descendant_of(fresh_root_node))
        self.assertFalse(fresh_root_node.is_descendant_of(fresh_bottom_node))

        self.assertTrue(fresh_bottom_node.is_descendant_of(fresh_middle_node))
        self.assertFalse(fresh_middle_node.is_descendant_of(fresh_bottom_node))

        self.assertFalse(fresh_root_node.is_descendant_of(fresh_root_node))
        self.assertFalse(fresh_middle_node.is_descendant_of(fresh_middle_node))
        self.assertFalse(fresh_bottom_node.is_descendant_of(fresh_bottom_node))

    def test_tree_structure_is_descendant_of_none(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)

        fresh_root_node = SimpleNode.objects.get(id=root_node.id)
        fresh_middle_node = SimpleNode.objects.get(id=middle_node.id)

        self.assertTrue(fresh_middle_node.is_descendant_of(fresh_root_node))

        self.assertFalse(fresh_middle_node.is_descendant_of(None))

    def test_tree_structure_is_ancestor_of(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node = SimpleNode.objects.create(parent=middle_node)

        fresh_root_node = SimpleNode.objects.get(id=root_node.id)
        fresh_middle_node = SimpleNode.objects.get(id=middle_node.id)
        fresh_bottom_node = SimpleNode.objects.get(id=bottom_node.id)

        self.assertFalse(fresh_middle_node.is_ancestor_of(fresh_root_node))
        self.assertTrue(fresh_root_node.is_ancestor_of(fresh_middle_node))

        self.assertFalse(fresh_bottom_node.is_ancestor_of(fresh_root_node))
        self.assertTrue(fresh_root_node.is_ancestor_of(fresh_bottom_node))

        self.assertFalse(fresh_bottom_node.is_ancestor_of(fresh_middle_node))
        self.assertTrue(fresh_middle_node.is_ancestor_of(fresh_bottom_node))

        self.assertFalse(fresh_root_node.is_ancestor_of(fresh_root_node))
        self.assertFalse(fresh_middle_node.is_ancestor_of(fresh_middle_node))
        self.assertFalse(fresh_bottom_node.is_ancestor_of(fresh_bottom_node))

    def test_tree_structure_is_ancestor_of_none(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)

        fresh_root_node = SimpleNode.objects.get(id=root_node.id)
        fresh_middle_node = SimpleNode.objects.get(id=middle_node.id)

        self.assertTrue(fresh_root_node.is_ancestor_of(fresh_middle_node))
        self.assertFalse(fresh_root_node.is_ancestor_of(None))

    def test_tree_structure_is_leaf(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node_1 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_2 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_3 = SimpleNode.objects.create(parent=middle_node)

        self.assertFalse(root_node.is_leaf())
        self.assertFalse(middle_node.is_leaf())
        self.assertTrue(bottom_node_1.is_leaf())
        self.assertTrue(bottom_node_2.is_leaf())
        self.assertTrue(bottom_node_3.is_leaf())

    def test_tree_structure_is_branch(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node_1 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_2 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_3 = SimpleNode.objects.create(parent=middle_node)

        self.assertTrue(root_node.is_branch())
        self.assertTrue(middle_node.is_branch())
        self.assertFalse(bottom_node_1.is_branch())
        self.assertFalse(bottom_node_2.is_branch())
        self.assertFalse(bottom_node_3.is_branch())

    def test_tree_attribute_path(self):

        root_node = SimpleNamedNode.objects.create(name='root')
        middle_node = SimpleNamedNode.objects.create(parent=root_node,
                                                     name='middle')
        SimpleNamedNode.objects.create(parent=middle_node, name='bottom')

        fresh_root_node = SimpleNamedNode.objects.get(name='root')
        fresh_middle_node = SimpleNamedNode.objects.get(name='middle')
        fresh_bottom_node = SimpleNamedNode.objects.get(name='bottom')

        self.assertEqual(['root', ], fresh_root_node.attribute_path('name'))
        self.assertEqual(['root', 'middle', ],
                         fresh_middle_node.attribute_path('name'))
        self.assertEqual(['root', 'middle', 'bottom', ],
                         fresh_bottom_node.attribute_path('name'))

        self.assertEqual(['foo', ], fresh_root_node.attribute_path('bar',
                                                                   missing='foo'))

        def custom_visitor(node, attribute):
            value = getattr(node, attribute, None)
            if value == 'middle':
                return 'foo'
            if value == 'bottom':
                return None
            return value

        self.assertEqual(['root', 'foo', 'bar', ],
                         fresh_bottom_node.attribute_path('name', missing='bar',
                                                          visitor=custom_visitor))

    def test_tree_structure_as_tree(self):

        root_node = OrderedNamedNode.objects.create(name='root')
        middle_node = OrderedNamedNode.objects.create(parent=root_node,
                                                      name='middle')
        OrderedNamedNode.objects.create(parent=middle_node, name='bottom 3')
        OrderedNamedNode.objects.create(parent=middle_node, name='bottom 1')
        OrderedNamedNode.objects.create(parent=middle_node, name='bottom 2')
        OrderedNamedNode.objects.create(name='root other')

        OrderedNamedNode.objects.get(name='root')
        fresh_middle_node = OrderedNamedNode.objects.get(name='middle')
        OrderedNamedNode.objects.get(name='bottom 1')
        fresh_bottom_node_2 = OrderedNamedNode.objects.get(name='bottom 2')
        OrderedNamedNode.objects.get(name='bottom 3')
        OrderedNamedNode.objects.get(name='root other')

        # get the forest
        tree = OrderedNamedNode.objects.as_tree()
        self.assertEqual(len(tree), 2)
        self.assertEqual(len(tree[0]['children']), 1)
        self.assertEqual(len(tree[1]['children']), 0)
        self.assertEqual(tree[0]['children'][0]['children'][1]['node'].id,
                         fresh_bottom_node_2.id)

        # get a specific tree
        middle_tree = fresh_middle_node.as_tree()
        self.assertEqual(middle_tree['node'].id, middle_node.id)
        self.assertFalse(middle_tree['leaf'])
        self.assertTrue(middle_tree['branch'])
        self.assertEqual(middle_tree['depth'], 2)
        self.assertEqual(middle_tree['ordering'], ['root', 'middle', ])
        self.assertEqual(middle_tree['path'],
                         [str(root_node.id), str(middle_node.id), ])
        self.assertEqual(len(middle_tree['children']), 3)
        self.assertEqual(middle_tree['children'][1]['node'].id,
                         fresh_bottom_node_2.id)

    def test_tree_structure_as_tree_custom(self):

        root_node = OrderedNamedNode.objects.create(name='root')
        middle_node = OrderedNamedNode.objects.create(parent=root_node,
                                                      name='middle')
        OrderedNamedNode.objects.create(parent=middle_node, name='bottom 3')
        OrderedNamedNode.objects.create(parent=middle_node, name='bottom 1')
        OrderedNamedNode.objects.create(parent=middle_node, name='bottom 2')
        OrderedNamedNode.objects.create(name='root other')

        OrderedNamedNode.objects.get(name='root')
        fresh_middle_node = OrderedNamedNode.objects.get(name='middle')
        OrderedNamedNode.objects.get(name='bottom 1')
        OrderedNamedNode.objects.get(name='bottom 2')
        fresh_bottom_node_3 = OrderedNamedNode.objects.get(name='bottom 3')
        OrderedNamedNode.objects.get(name='root other')

        def custom_visitor(manager, node):
            tree = manager._default_node_visitor(node)
            tree['person'] = tree['node']
            del tree['node']
            return tree

        def custom_children(manager, node, visitor, children):
            if node.name == 'bottom 2':
                return {}
            return {'offspring': [
                child.as_tree(
                    visitor=visitor,
                    children=children,
                ) for child in node.children.exclude(name='bottom 2')
            ]}

        # get the forest
        tree = OrderedNamedNode.objects.as_tree(visitor=custom_visitor,
                                                children=custom_children)
        self.assertEqual(len(tree), 2)
        self.assertEqual(len(tree[0]['offspring']), 1)
        self.assertEqual(len(tree[1]['offspring']), 0)
        self.assertEqual(tree[0]['offspring'][0]['offspring'][1]['person'].id,
                         fresh_bottom_node_3.id)

        # get a specific tree
        middle_tree = fresh_middle_node.as_tree(visitor=custom_visitor,
                                                children=custom_children)
        self.assertEqual(middle_tree['person'].id, middle_node.id)
        self.assertFalse(middle_tree['leaf'])
        self.assertTrue(middle_tree['branch'])
        self.assertEqual(middle_tree['depth'], 2)
        self.assertEqual(middle_tree['ordering'], ['root', 'middle', ])
        self.assertEqual(middle_tree['path'],
                         [str(root_node.id), str(middle_node.id), ])
        self.assertEqual(len(middle_tree['offspring']), 2)
        self.assertEqual(middle_tree['offspring'][1]['person'].id,
                         fresh_bottom_node_3.id)

    def test_tree_drilldown(self):

        # check empty forest (no roots, empty path)
        self.assertRaises(SimpleNamedNode.DoesNotExist,
                          lambda: SimpleNamedNode.objects.drilldown(('name', ), []))

        root_node = SimpleNamedNode.objects.create(name='root')
        middle_node = SimpleNamedNode.objects.create(parent=root_node,
                                                     name='middle')
        SimpleNamedNode.objects.create(parent=middle_node, name='bottom')

        fresh_root_node = SimpleNamedNode.objects.get(name='root')
        SimpleNamedNode.objects.get(name='middle')
        fresh_bottom_node = SimpleNamedNode.objects.get(name='bottom')

        # check missing path component
        self.assertRaises(
            SimpleNamedNode.DoesNotExist,
            lambda: SimpleNamedNode.objects.drilldown(('name', ), [('root', ), ('xxx')]))

        # check missing root component
        self.assertRaises(
            SimpleNamedNode.DoesNotExist,
            lambda: SimpleNamedNode.objects.drilldown(('name', ), [('xxx', ), ('xxx')]))

        # check success
        self.assertEqual(fresh_bottom_node, SimpleNamedNode.objects.drilldown(
            ('name', ), [('root', ), ('middle', ), ('bottom', )]))

        # check empty path success
        self.assertEqual(
            fresh_root_node,
            SimpleNamedNode.objects.drilldown(('name',), []))

        # check extraneous path component
        self.assertRaises(
            SimpleNamedNode.DoesNotExist,
            lambda: SimpleNamedNode.objects.drilldown(
                ('name',),
                [('root',), ('middle',), ('bottom',), ('xxx',)],
            ),
        )

    def test_tree_drilldown_complex_filtering(self):

        root_node = ValueNamedNode.objects.create(name='root', v=5)
        middle_node = ValueNamedNode.objects.create(
            parent=root_node, name='middle', v=5)
        ValueNamedNode.objects.create(
            parent=middle_node, name='xxx bottom 1', v=7)
        ValueNamedNode.objects.create(
            parent=middle_node, name='bottom 2', v=1)
        ValueNamedNode.objects.create(parent=middle_node, name='bottom 3', v=6)

        ValueNamedNode.objects.get(name='root')
        ValueNamedNode.objects.get(name='middle')
        ValueNamedNode.objects.get(name='xxx bottom 1')
        ValueNamedNode.objects.get(name='bottom 2')
        fresh_bottom_node_3 = ValueNamedNode.objects.get(name='bottom 3')

        self.assertEqual(fresh_bottom_node_3, ValueNamedNode.objects.drilldown(
            ('name__startswith', 'v__gte'),
            [('root', 5), ('middle', 5), ('bottom', 5)]))

    def test_node_delete_pharaoh(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        SimpleNode.objects.create(parent=middle_node)
        SimpleNode.objects.create(parent=middle_node)
        SimpleNode.objects.create(parent=middle_node)

        middle_node.delete()

        self.assertEqual(len(SimpleNode.objects.all()), 1)
        self.assertEqual(SimpleNode.objects.count(), 1)
        self.assertEqual(SimpleNode.objects.get().id, root_node.id)

    def test_node_delete_grandmother(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node_1 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_2 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_3 = SimpleNode.objects.create(parent=middle_node)

        middle_node.delete(
            method=SimpleNode.objects.DELETE_METHOD_GRANDMOTHER)

        self.assertEqual(len(SimpleNode.objects.all()), 4)
        self.assertEqual(SimpleNode.objects.count(), 4)
        self.assertEqual([node.id for node in SimpleNode.objects.all()],
                         [root_node.id, bottom_node_1.id, bottom_node_2.id,
                          bottom_node_3.id])
        self.assertEqual(len(root_node.children.all()), 3)
        self.assertEqual(root_node.children.count(), 3)
        self.assertEqual([node.depth for node in root_node.children.all()],
                         [2, 2, 2])

    def test_node_delete_monarchy(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        bottom_node_1 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_2 = SimpleNode.objects.create(parent=middle_node)
        bottom_node_3 = SimpleNode.objects.create(parent=middle_node)

        middle_node.delete(method=SimpleNode.objects.DELETE_METHOD_MONARCHY)

        fresh_bottom_node_1 = SimpleNode.objects.get(id=bottom_node_1.id)

        self.assertEqual(len(SimpleNode.objects.all()), 4)
        self.assertEqual(SimpleNode.objects.count(), 4)
        self.assertEqual([node.id for node in SimpleNode.objects.all()],
                         [root_node.id, bottom_node_1.id, bottom_node_2.id,
                          bottom_node_3.id])
        self.assertEqual(len(root_node.children.all()), 1)
        self.assertEqual(root_node.children.count(), 1)
        self.assertEqual([node.depth for node in root_node.children.all()], [2])

        self.assertEqual(fresh_bottom_node_1.depth, 2)
        self.assertEqual(fresh_bottom_node_1.children.count(), 2)
        self.assertEqual(len(fresh_bottom_node_1.children.all()), 2)
        self.assertEqual([3, 3],
                         [node.depth for node in fresh_bottom_node_1.children.all()])

    def test_node_delete_none(self):

        root_node = NoneDeleteNode.objects.create()
        middle_node = NoneDeleteNode.objects.create(parent=root_node)
        NoneDeleteNode.objects.create(parent=middle_node)
        NoneDeleteNode.objects.create(parent=middle_node)
        NoneDeleteNode.objects.create(parent=middle_node)

        middle_node.delete()

        self.assertEqual(len(NoneDeleteNode.objects.all()), 1)
        self.assertEqual(NoneDeleteNode.objects.count(), 1)
        self.assertEqual(NoneDeleteNode.objects.get().id, root_node.id)

    def test_node_move(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        SimpleNode.objects.create(parent=middle_node)
        SimpleNode.objects.create(parent=middle_node)
        bottom_node_3 = SimpleNode.objects.create(parent=middle_node)

        fresh_bottom_node_3 = SimpleNode.objects.get(id=bottom_node_3.id)

        self.assertEqual(fresh_bottom_node_3.depth, 3)

        bottom_node_3.move(root_node, position=lambda node, destination: None,
                           save=True)

        fresh_bottom_node_3 = SimpleNode.objects.get(id=bottom_node_3.id)

        self.assertEqual(fresh_bottom_node_3.depth, 2)


class SimpleNodeErrorsTest(TestCase):

    def test_node_parameters_parent_1(self):

        # _cte_node_parent points to a non-existing field.
        self.assertRaises(ImproperlyConfigured,
                          BadParameter_parent_1_Node.objects.create)

    def test_node_parameters_parent_2(self):

        # Parent Foreign Key points to different CTENode Model.
        self.assertRaises(ImproperlyConfigured,
                          BadParameter_parent_2_Node.objects.create)

    def test_node_parameters_parent_3(self):

        # Parent Foreign Key points to arbitrary Model.
        self.assertRaises(ImproperlyConfigured,
                          BadParameter_parent_3_Node.objects.create)

    def test_node_parameters_parent_4(self):

        # Parent Foreign Key missing.
        self.assertRaises(ImproperlyConfigured,
                          BadParameter_parent_4_Node.objects.create)

    def test_node_parameters_traversal(self):

        self.assertRaises(ImproperlyConfigured,
                          BadParameter_traversal_Node.objects.create)

    def test_node_parameters_delete(self):

        self.assertRaises(ImproperlyConfigured,
                          BadParameter_delete_Node.objects.create)

    def test_node_virtual_fields(self):

        root_node = SimpleNode.objects.create()

        def read_path(node):
            return node.path

        self.assertRaises(AttributeError, read_path, root_node)
        self.assertRaises(FieldError, root_node.ancestors)


class SimpleNodeUsageTest(TestCase):

    def test_node_usage(self):

        root_node = SimpleNode.objects.create()
        middle_node = SimpleNode.objects.create(parent=root_node)
        SimpleNode.objects.create(parent=middle_node)
        SimpleNode.objects.create(parent=middle_node)
        SimpleNode.objects.create(parent=middle_node)

        SimpleNodeUser.objects.create(node=root_node)

        self.assertEqual(SimpleNodeUser.objects.get().node.id, root_node.id)


class SimpleNamedNodeTest(TestCase):

    def test_node_creation(self):

        node = SimpleNamedNode.objects.create(name='root')

        self.assertEqual(SimpleNamedNode.objects.get().name, node.name)

    def test_user_creation(self):

        node = SimpleNamedNode.objects.create(name='root')
        user = SimpleNamedNodeUser.objects.create(node=node,
                                                  name='root user')

        self.assertEqual(user.node.name, 'root')

    def test_node_save(self):

        node = SimpleNamedNode.objects.create(name='amazing')
        self.assertEqual(SimpleNamedNode.objects.get().name, 'amazing')
        node.name = 'so and so'
        node.save()
        self.assertEqual(SimpleNamedNode.objects.get().name, 'so and so')

    def test_ordering(self):

        root_node = SimpleNamedNode.objects.create(name='root')
        middle_node = SimpleNamedNode.objects.create(parent=root_node,
                                                     name='middle')

        # Create these in mixed order to test ordering by name below.
        SimpleNamedNode.objects.create(parent=middle_node, name='bottom 2')
        SimpleNamedNode.objects.create(parent=middle_node, name='bottom 3')
        SimpleNamedNode.objects.create(parent=middle_node, name='bottom 1')

        # Order should be by primary key, not name.
        node_names = [node.name for node in SimpleNamedNode.objects.all()]
        self.assertEqual(node_names,
                         ['root', 'middle', 'bottom 2', 'bottom 3', 'bottom 1'])

        # But if we override ordering, we can get back to flat name space.
        flat_node_names = [
            node.name for node in
            SimpleNamedNode.objects.all().order_by('name')]
        self.assertEqual(flat_node_names,
                         ['bottom 1', 'bottom 2', 'bottom 3', 'middle', 'root'])

    def test_node_filter_chaining(self):

        root_node = SimpleNamedNode.objects.create(name='root')
        middle_node = SimpleNamedNode.objects.create(parent=root_node,
                                                     name='middle')
        SimpleNamedNode.objects.create(parent=middle_node, name='bottom 1')
        SimpleNamedNode.objects.create(parent=middle_node, name='bottom 2')
        SimpleNamedNode.objects.create(parent=middle_node, name='bottom 3')

        self.assertEqual(
            'bottom 1',
            SimpleNamedNode.objects.filter(
                id__gt=1,
            ).exclude(
                name='bottom 3',
            ).filter(
                name__in=['bottom 3', 'bottom 1'],
            )[0].name,
        )


class OrderedNamedNodeTest(TestCase):

    def test_node_creation(self):

        node = OrderedNamedNode.objects.create(name='root')
        fresh_node = OrderedNamedNode.objects.get()

        self.assertEqual(fresh_node.ordering, [node.name])

    def test_simple_ordering(self):

        root_node = OrderedNamedNode.objects.create(name='root')
        middle_node = OrderedNamedNode.objects.create(parent=root_node,
                                                      name='middle')

        # Create these in mixed order to test ordering by name below.
        OrderedNamedNode.objects.create(parent=middle_node, name='bottom 2')
        OrderedNamedNode.objects.create(parent=middle_node, name='bottom 3')
        OrderedNamedNode.objects.create(parent=middle_node, name='bottom 1')

        # Order should be by path name, not primary key.
        node_names = [node.name for node in OrderedNamedNode.objects.all()]
        self.assertEqual(['root', 'middle', 'bottom 1', 'bottom 2', 'bottom 3'],
                         node_names)

        # But if we override ordering, we can get back to flat name space.
        flat_node_names = [
            node.name for node in
            OrderedNamedNode.objects.all().order_by('name')]
        self.assertEqual(['bottom 1', 'bottom 2', 'bottom 3', 'middle', 'root'],
                         flat_node_names)


class DFSOrderedNodeTest(TestCase):

    def test_ordering_dfs(self):

        root_node = DFSOrderedNode.objects.create(v=1)
        # The following two have different v values, but created in the
        # opposite order from which we are ordering.
        middle_node_1 = DFSOrderedNode.objects.create(parent=root_node,
                                                      v=5)
        middle_node_2 = DFSOrderedNode.objects.create(parent=root_node,
                                                      v=4)

        # The following two have the same v value.
        bottom_node_1_1 = DFSOrderedNode.objects.create(
            parent=middle_node_1, v=3)
        bottom_node_1_2 = DFSOrderedNode.objects.create(
            parent=middle_node_1, v=3)

        # The following have different v values, but created in 'reverse' order.
        bottom_node_2_1 = DFSOrderedNode.objects.create(
            parent=middle_node_2, v=5)
        bottom_node_2_2 = DFSOrderedNode.objects.create(
            parent=middle_node_2, v=4)

        expected_order = [root_node.id, middle_node_2.id, bottom_node_2_2.id,
                          bottom_node_2_1.id, middle_node_1.id, bottom_node_1_1.id,
                          bottom_node_1_2.id]

        self.assertEqual(expected_order,
                         [node.id for node in DFSOrderedNode.objects.all()])


class BFSOrderedNodeTest(TestCase):

    def test_ordering_bfs(self):

        root_node = BFSOrderedNode.objects.create(v=1)
        # The following two have different v values, but created in the
        # opposite order from which we are ordering.
        middle_node_1 = BFSOrderedNode.objects.create(parent=root_node,
                                                      v=5)
        middle_node_2 = BFSOrderedNode.objects.create(parent=root_node,
                                                      v=4)

        # The following two have the same v value.
        bottom_node_1_1 = BFSOrderedNode.objects.create(
            parent=middle_node_1, v=3)
        bottom_node_1_2 = BFSOrderedNode.objects.create(
            parent=middle_node_1, v=3)

        # The following have different v values, but created in 'reverse' order.
        bottom_node_2_1 = BFSOrderedNode.objects.create(
            parent=middle_node_2, v=5)
        bottom_node_2_2 = BFSOrderedNode.objects.create(
            parent=middle_node_2, v=4)

        expected_order = [root_node.id, middle_node_2.id, middle_node_1.id,
                          bottom_node_2_2.id, bottom_node_2_1.id, bottom_node_1_1.id,
                          bottom_node_1_2.id]

        self.assertEqual(expected_order,
                         [node.id for node in BFSOrderedNode.objects.all()])


class TraversalNodeTest(TestCase):

    def test_none_traversal_parameter(self):

        root_node = NoneTraversalNode.objects.create(v=1)
        # The following two have different v values, but created in the
        # opposite order from which we are ordering.
        middle_node_1 = NoneTraversalNode.objects.create(parent=root_node,
                                                         v=5)
        middle_node_2 = NoneTraversalNode.objects.create(parent=root_node,
                                                         v=4)

        # The following two have the same v value.
        bottom_node_1_1 = NoneTraversalNode.objects.create(
            parent=middle_node_1, v=3)
        bottom_node_1_2 = NoneTraversalNode.objects.create(
            parent=middle_node_1, v=3)

        # The following have different v values, but created in 'reverse' order.
        bottom_node_2_1 = NoneTraversalNode.objects.create(
            parent=middle_node_2, v=5)
        bottom_node_2_2 = NoneTraversalNode.objects.create(
            parent=middle_node_2, v=4)

        expected_order = [root_node.id, middle_node_2.id, bottom_node_2_2.id,
                          bottom_node_2_1.id, middle_node_1.id, bottom_node_1_1.id,
                          bottom_node_1_2.id]

        self.assertEqual(expected_order,
                         [node.id for node in NoneTraversalNode.objects.all()])


class TypeCoercionNodeTest(TestCase):

    def test_type_coercion(self):

        # Note: this tests in DFS. We will order on a VARCHAR, which will be
        # automatically cast into TEXT, and also on an int, which we will have
        # to explicitly cast into TEXT in order to combine with the VARCHAR.

        root_node = TypeCoercionNode.objects.create(v=1, name='')
        # The following two have different v values, but created in the
        # opposite order from which we are ordering.
        middle_node_1 = TypeCoercionNode.objects.create(parent=root_node,
                                                        v=5, name='foo')
        middle_node_2 = TypeCoercionNode.objects.create(parent=root_node,
                                                        v=4, name='foo')

        # The following two have the same v value.
        bottom_node_1_1 = TypeCoercionNode.objects.create(
            parent=middle_node_1, v=3, name='b')
        bottom_node_1_2 = TypeCoercionNode.objects.create(
            parent=middle_node_1, v=3, name='a')

        # The following two have the same v and name values, so the order in
        # they will be returned is defined by the database.
        bottom_node_2_1 = TypeCoercionNode.objects.create(
            parent=middle_node_2, v=4, name='foo')
        bottom_node_2_2 = TypeCoercionNode.objects.create(
            parent=middle_node_2, v=4, name='foo')

        expected_order = [root_node.id, middle_node_2.id, bottom_node_2_2.id,
                          bottom_node_2_1.id, middle_node_1.id, bottom_node_1_2.id,
                          bottom_node_1_1.id]

        self.assertEqual(expected_order,
                         [node.id for node in TypeCoercionNode.objects.all()])


class TypeCombinationNodeTest(TestCase):

    def test_type_combination(self):

        # Note: this tests in BFS.

        root_node = TypeCombinationNode.objects.create(v1=1, v2=3.2)

        # The following two have different v1 values, but created in the
        # opposite order from which we are ordering.
        middle_node_1 = TypeCombinationNode.objects.create(parent=root_node,
                                                           v1=5, v2=3.7)
        middle_node_2 = TypeCombinationNode.objects.create(parent=root_node,
                                                           v1=4, v2=8.9)

        # The following two have the same v1 value.
        bottom_node_1_1 = TypeCombinationNode.objects.create(
            parent=middle_node_1, v1=3, v2=1.5)
        bottom_node_1_2 = TypeCombinationNode.objects.create(
            parent=middle_node_1, v1=3, v2=1.4)

        # The following two have the same v1 and v2 values, so the order in
        # they will be returned is defined by the database.
        bottom_node_2_1 = TypeCombinationNode.objects.create(
            parent=middle_node_2, v1=4, v2=4)
        bottom_node_2_2 = TypeCombinationNode.objects.create(
            parent=middle_node_2, v1=4, v2=5)

        expected_order = [root_node.id, middle_node_2.id, middle_node_1.id,
                          bottom_node_2_1.id, bottom_node_2_2.id, bottom_node_1_2.id,
                          bottom_node_1_1.id]

        self.assertEqual(expected_order,
                         [node.id for node in TypeCombinationNode.objects.all()])


class ExoticTypeNodeTest(TestCase):

    def test_exotic_type(self):

        root_node = ExoticTypeNode.objects.create(v=date(1982, 9, 26))
        # The following two have different v values, but created in the
        # opposite order from which we are ordering.
        middle_node_1 = ExoticTypeNode.objects.create(parent=root_node,
                                                      v=date(2006, 7, 7))
        middle_node_2 = ExoticTypeNode.objects.create(parent=root_node,
                                                      v=date(1946, 1, 6))

        # The following two have the same v value.
        bottom_node_1_1 = ExoticTypeNode.objects.create(
            parent=middle_node_1, v=date(1924, 11, 20))
        bottom_node_1_2 = ExoticTypeNode.objects.create(
            parent=middle_node_1, v=date(1924, 11, 20))

        # The following have different v values, but created in 'reverse' order.
        bottom_node_2_1 = ExoticTypeNode.objects.create(
            parent=middle_node_2, v=date(1987, 10, 20))
        bottom_node_2_2 = ExoticTypeNode.objects.create(
            parent=middle_node_2, v=date(1903, 4, 25))

        expected_order = [root_node.id, middle_node_2.id, bottom_node_2_2.id,
                          bottom_node_2_1.id, middle_node_1.id, bottom_node_1_1.id,
                          bottom_node_1_2.id]

        self.assertEqual(expected_order,
                         [node.id for node in ExoticTypeNode.objects.all()])

    def test_date_query(self):

        node1 = ExoticTypeNode.objects.create(v=date(1982, 9, 26),
                                              y=date(1982, 9, 29))
        node2 = ExoticTypeNode.objects.create(v=date(1982, 9, 26),
                                              y=date(1982, 9, 30))

        self.assertEqual(list(ExoticTypeNode.objects.filter(
            y__gt=F('v') + timedelta(days=3))), [node2])

        self.assertEqual(ExoticTypeNode.objects.filter(
            y__gt=F('v') + timedelta(days=3)).count(), 1)

        self.assertEqual(ExoticTypeNode.objects.filter(
            y__gte=F('v') + timedelta(days=3)).count(), 2)

        self.assertEqual(list(ExoticTypeNode.objects.filter(
            y__lte=F('v') + timedelta(days=3))), [node1])

        self.assertEqual(list(ExoticTypeNode.objects.filter(
            y__lte=F('v') + timedelta(days=-4))), [])


class DBTypeNodeTest(TestCase):

    sorted_uuids = [
        UUID('24f8aab3-43b1-4359-bbe7-532e33aa8114'),
        UUID('4761ac94-9435-4fe0-9644-3e0b0da4e808'),
        UUID('563a973c-74da-46f9-8be7-70b9b60ca47a'),
        UUID('6309a02a-fec0-48c3-a93d-4a06a80de78b'),
        UUID('75025ae9-fd6d-41d2-8cd3-c34c49267d6d'),
        UUID('78e66b6f-f59c-42df-b20e-ea9855329bea')
    ]

    def test_ensure_sorted_uuids(self):

        previous = None
        for u in self.sorted_uuids:
            if previous is None:
                previous = u
            else:
                self.assertTrue(previous < u)

    def test_db_type(self):

        root_node = DBTypeNode.objects.create(v=self.sorted_uuids[0])
        # The following two have different v values, but created in the
        # opposite order from which we are ordering.
        middle_node_1 = DBTypeNode.objects.create(parent=root_node,
                                                  v=self.sorted_uuids[2])
        middle_node_2 = DBTypeNode.objects.create(parent=root_node,
                                                  v=self.sorted_uuids[1])

        # The following two have the same v value.
        bottom_node_1_1 = DBTypeNode.objects.create(
            parent=middle_node_1, v=self.sorted_uuids[3])
        bottom_node_1_2 = DBTypeNode.objects.create(
            parent=middle_node_1, v=self.sorted_uuids[3])

        # The following have different v values, but created in 'reverse' order.
        bottom_node_2_1 = DBTypeNode.objects.create(
            parent=middle_node_2, v=self.sorted_uuids[5])
        bottom_node_2_2 = DBTypeNode.objects.create(
            parent=middle_node_2, v=self.sorted_uuids[4])

        expected_order = [root_node.id, middle_node_2.id, bottom_node_2_2.id,
                          bottom_node_2_1.id, middle_node_1.id, bottom_node_1_1.id,
                          bottom_node_1_2.id]

        self.assertEqual(expected_order,
                         [node.id for node in DBTypeNode.objects.all()])

    def test_db_type_path(self):

        root_node = DBTypeNode.objects.create(v=self.sorted_uuids[0])
        # The following two have different v values, but created in the
        # opposite order from which we are ordering.
        middle_node_1 = DBTypeNode.objects.create(parent=root_node,
                                                  v=self.sorted_uuids[2])
        DBTypeNode.objects.create(parent=root_node, v=self.sorted_uuids[1])

        fresh_middle_node_1 = DBTypeNode.objects.get(id=middle_node_1.id)

        self.assertEqual(fresh_middle_node_1.ancestors()[0], root_node)

    def test_tree_structure_is_child_of(self):

        root_node = DBTypeNode.objects.create()
        middle_node = DBTypeNode.objects.create(parent=root_node)
        bottom_node_1 = DBTypeNode.objects.create(parent=middle_node)
        DBTypeNode.objects.create(parent=middle_node)
        DBTypeNode.objects.create(parent=middle_node)

        fresh_middle_node = DBTypeNode.objects.get(id=middle_node.id)
        fresh_bottom_node_1 = DBTypeNode.objects.get(id=bottom_node_1.id)

        self.assertTrue(fresh_middle_node.is_child_of(root_node))
        self.assertFalse(fresh_bottom_node_1.is_child_of(root_node))


class CustomPrimaryKeyNodeTest(TestCase):

    def test_tree_structure_is_child_of(self):

        # Ensure string-based path encoding works.

        root_node = CustomPrimaryKeyNode.objects.create(id='root')
        middle_node = CustomPrimaryKeyNode.objects.create(id='middle',
                                                          parent=root_node)
        bottom_node_1 = CustomPrimaryKeyNode.objects.create(id='bottom 1',
                                                            parent=middle_node)
        CustomPrimaryKeyNode.objects.create(id='bottom 2', parent=middle_node)
        CustomPrimaryKeyNode.objects.create(id='bottom 3', parent=middle_node)

        fresh_middle_node = CustomPrimaryKeyNode.objects.get(
            id=middle_node.id)
        fresh_bottom_node_1 = CustomPrimaryKeyNode.objects.get(
            id=bottom_node_1.id)

        self.assertTrue(fresh_middle_node.is_child_of(root_node))
        self.assertFalse(fresh_bottom_node_1.is_child_of(root_node))

    def test_tree_structure_is_descendant_of(self):

        # Ensure string-based path encoding works.

        root_node = CustomPrimaryKeyNode.objects.create(id='root')
        middle_node = CustomPrimaryKeyNode.objects.create(id='middle',
                                                          parent=root_node)
        bottom_node_1 = CustomPrimaryKeyNode.objects.create(id='bottom 1',
                                                            parent=middle_node)
        CustomPrimaryKeyNode.objects.create(id='bottom 2', parent=middle_node)
        CustomPrimaryKeyNode.objects.create(id='bottom 3', parent=middle_node)

        fresh_root_node = CustomPrimaryKeyNode.objects.get(id=root_node.id)
        fresh_middle_node = CustomPrimaryKeyNode.objects.get(
            id=middle_node.id)
        fresh_bottom_node_1 = CustomPrimaryKeyNode.objects.get(
            id=bottom_node_1.id)

        self.assertTrue(fresh_bottom_node_1.is_descendant_of(fresh_root_node))

        self.assertTrue(fresh_root_node.is_ancestor_of(fresh_bottom_node_1))

        self.assertFalse(fresh_middle_node.is_descendant_of(
            fresh_bottom_node_1))

        self.assertFalse(fresh_bottom_node_1.is_ancestor_of(fresh_middle_node))


class DBTypePrimaryKeyNodeTest(TestCase):

    def test_tree_structure_is_child_of(self):

        # Ensure string-based path encoding works.

        root_node = DBTypePrimaryKeyNode.objects.create()
        middle_node = DBTypePrimaryKeyNode.objects.create(parent=root_node)
        bottom_node_1 = DBTypePrimaryKeyNode.objects.create(
            parent=middle_node)
        DBTypePrimaryKeyNode.objects.create(
            parent=middle_node)
        DBTypePrimaryKeyNode.objects.create(
            parent=middle_node)

        fresh_middle_node = DBTypePrimaryKeyNode.objects.get(
            id=middle_node.id)
        fresh_bottom_node_1 = DBTypePrimaryKeyNode.objects.get(
            id=bottom_node_1.id)

        self.assertTrue(fresh_middle_node.is_child_of(root_node))
        self.assertFalse(fresh_bottom_node_1.is_child_of(root_node))

    def test_tree_structure_is_ancestor_of(self):

        # Ensure string-based path encoding works.

        root_node = DBTypePrimaryKeyNode.objects.create()
        middle_node = DBTypePrimaryKeyNode.objects.create(parent=root_node)
        bottom_node = DBTypePrimaryKeyNode.objects.create(parent=middle_node)

        fresh_root_node = DBTypePrimaryKeyNode.objects.get(id=root_node.id)
        fresh_middle_node = DBTypePrimaryKeyNode.objects.get(
            id=middle_node.id)
        fresh_bottom_node = DBTypePrimaryKeyNode.objects.get(
            id=bottom_node.id)

        self.assertFalse(fresh_middle_node.is_ancestor_of(fresh_root_node))
        self.assertTrue(fresh_root_node.is_ancestor_of(fresh_middle_node))

        self.assertFalse(fresh_bottom_node.is_ancestor_of(fresh_root_node))
        self.assertTrue(fresh_root_node.is_ancestor_of(fresh_bottom_node))

        self.assertFalse(fresh_bottom_node.is_ancestor_of(fresh_middle_node))
        self.assertTrue(fresh_middle_node.is_ancestor_of(fresh_bottom_node))

        self.assertFalse(fresh_root_node.is_ancestor_of(fresh_root_node))
        self.assertFalse(fresh_middle_node.is_ancestor_of(fresh_middle_node))
        self.assertFalse(fresh_bottom_node.is_ancestor_of(fresh_bottom_node))

    def test_tree_structure_is_descendant_of(self):

        # Ensure string-based path encoding works.

        root_node = DBTypePrimaryKeyNode.objects.create()
        middle_node = DBTypePrimaryKeyNode.objects.create(parent=root_node)
        bottom_node = DBTypePrimaryKeyNode.objects.create(parent=middle_node)

        fresh_root_node = DBTypePrimaryKeyNode.objects.get(id=root_node.id)
        fresh_middle_node = DBTypePrimaryKeyNode.objects.get(
            id=middle_node.id)
        fresh_bottom_node = DBTypePrimaryKeyNode.objects.get(
            id=bottom_node.id)

        self.assertTrue(fresh_middle_node.is_descendant_of(fresh_root_node))
        self.assertFalse(fresh_root_node.is_descendant_of(fresh_middle_node))

        self.assertTrue(fresh_bottom_node.is_descendant_of(fresh_root_node))
        self.assertFalse(fresh_root_node.is_descendant_of(fresh_bottom_node))

        self.assertTrue(fresh_bottom_node.is_descendant_of(fresh_middle_node))
        self.assertFalse(fresh_middle_node.is_descendant_of(fresh_bottom_node))

        self.assertFalse(fresh_root_node.is_descendant_of(fresh_root_node))
        self.assertFalse(fresh_middle_node.is_descendant_of(fresh_middle_node))
        self.assertFalse(fresh_bottom_node.is_descendant_of(fresh_bottom_node))


class AggregationNodeTest(TestCase):

    def test_aggregation(self):

        AggregationNode.objects.create(price=100)
        node2 = AggregationNode.objects.create(price=200)
        AggregationNode.objects.create(price=300)

        self.assertEqual(
            AggregationNode.objects.all().aggregate(Avg('price'))['price__avg'],
            200)

        self.assertEqual(list(AggregationNode.objects.filter(
            price=AggregationNode.objects.all().aggregate(
                Avg('price'))['price__avg'])), [node2])



# -*- coding: utf-8 -*-
#
# This document is free and open-source software, subject to the OSI-approved
# BSD license below.
#
# Copyright (c) 2011 - 2013 Alexis Petrounias <www.petrounias.org>,
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the author nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

""" Django CTE Trees Query Compiler.
"""
from __future__ import unicode_literals

import django
from django.db import connections
from django.db.models.query import QuerySet
from django.db.models.sql import (
    AggregateQuery, DeleteQuery, InsertQuery, UpdateQuery,
)
from django.db.models.sql.compiler import (
    SQLAggregateCompiler, SQLCompiler, SQLDeleteCompiler, SQLInsertCompiler,
    SQLUpdateCompiler,
)
from django.db.models.sql.query import Query
from django.db.models.sql.where import ExtraWhere, WhereNode


class CTEQuerySet(QuerySet):
    """
    The QuerySet which ensures all CTE Node SQL compilation is processed by the
    CTE Compiler and has the appropriate extra syntax, selects, tables, and
    WHERE clauses.
    """

    def __init__(self, model=None, query=None, using=None, offset=None, hints=None):
        """
        Prepares a CTEQuery object by adding appropriate extras, namely the
        SELECT virtual fields, the WHERE clause which matches the CTE pk with
        the real table pk, and the tree-specific order_by parameters. If the
        query object has already been prepared through this phase, then it
        won't be prepared again.

        :param model:
        :type model:
        :param query:
        :type query:
        :param using:
        :type using:
        :param offset:
        :type offset:
        :return:
        :rtype:
        """
        # Only create an instance of a Query if this is the first invocation in
        # a query chain.
        if query is None:
            query = CTEQuery(model, offset=offset)
        super(CTEQuerySet, self).__init__(model, query, using)

    def aggregate(self, *args, **kwargs):
        """
        Returns a dictionary containing the calculations (aggregation)
        over the current queryset

        If args is present the expression is passed as a kwarg using
        the Aggregate object's default alias.
        """
        if self.query.distinct_fields:
            raise NotImplementedError("aggregate() + distinct(fields) not implemented.")
        for arg in args:
            # The default_alias property may raise a TypeError, so we use
            # a try/except construct rather than hasattr in order to remain
            # consistent between PY2 and PY3 (hasattr would swallow
            # the TypeError on PY2).
            try:
                arg.default_alias
            except (AttributeError, TypeError):
                raise TypeError("Complex aggregates require an alias")
            kwargs[arg.default_alias] = arg

        if django.VERSION < (2, 0):
            query = self.query.clone(CTEAggregateQuery)
        else:
            query = self.query.chain(CTEAggregateQuery)
        for (alias, aggregate_expr) in kwargs.items():
            query.add_annotation(aggregate_expr, alias, is_summary=True)
            if not query.annotations[alias].contains_aggregate:
                raise TypeError("%s is not an aggregate expression" % alias)
        return query.get_aggregation(self.db, kwargs.keys())


class CTEQuery(Query):
    """
    A Query which processes SQL compilation through the CTE Compiler, as well as
    keeps track of extra selects, the CTE table, and the corresponding WHERE
    clauses.
    """

    def __init__(self, model, where=WhereNode, offset=None):
        """
        Delegates initialization to Django's Query, and in addition adds the
        necessary extra selects, table, and where clauses for the CTE. Depending
        on the ordering semantics specified in the CTE Node as well as any
        order_by queries, the final order_by of the Query is generated. In the
        event of an offset Node being specified (such as querying for all of the
        descendants of a Node), an additional appropriate WHERE clauses is also
        added.

        :param model:
        :type model:
        :param where:
        :type where:
        :param offset:
        :type offset:
        :return:
        :rtype:
        """
        super(CTEQuery, self).__init__(model, where=where)
        # import from models here to avoid circular imports.
        from .models import CTENodeManager
        if model is not None:
            where = [self._generate_where(self)]
            # If an offset Node is specified, then only those Nodes which
            # contain the offset Node as a parent (in their path virtual field)
            # will be returned.
            if offset is not None:
                where.append("""'{id}' = ANY({cte}."{path}")""".format(
                    cte=model._cte_node_table,
                    path=model._cte_node_path,
                    id=str(offset.id)))
            order_by_prefix = []
            if model._cte_node_traversal == \
                    CTENodeManager.TREE_TRAVERSAL_NONE:
                chosen_traversal = CTENodeManager.DEFAULT_TREE_TRAVERSAL
            else:
                chosen_traversal = model._cte_node_traversal
            if chosen_traversal == CTENodeManager.TREE_TRAVERSAL_DFS:
                order_by_prefix = [model._cte_node_ordering]
            if chosen_traversal == CTENodeManager.TREE_TRAVERSAL_BFS:
                order_by_prefix = [model._cte_node_depth,
                                   model._cte_node_ordering]
            # prepend correct CTE table prefix to order_by fields
            order_by = [
                '{cte}.{field}'.format(cte=model._cte_node_table, field=field)
                for field in order_by_prefix]
            if hasattr(model, '_cte_node_order_by') and \
                    model._cte_node_order_by is not None and \
                    len(model._cte_node_order_by) > 0:
                order_by.extend(
                    field[0] if type(field) == tuple else field
                    for field in model._cte_node_order_by
                )
            # Specify the virtual fields for depth, path, and ordering;
            # optionally the offset Node constraint; and the desired ordering.
            # The CTE table will be added later by the Compiler only if the
            # actual query needs it.
            self.add_extra(
                select={
                    model._cte_node_depth: '{cte}.{depth}'.format(
                        cte=model._cte_node_table,
                        depth=model._cte_node_depth),
                    model._cte_node_path: '{cte}.{path}'.format(
                        cte=model._cte_node_table,
                        path=model._cte_node_path),
                    model._cte_node_ordering: '{cte}.{ordering}'.format(
                        cte=model._cte_node_table,
                        ordering=model._cte_node_ordering),
                },
                select_params=None,
                where=where,
                params=None,
                tables=None,
                order_by=order_by)

    @classmethod
    def _generate_where(cls, query):
        def maybe_alias(table):
            if table in query.table_map:
                return query.table_map[table][0]
            return table
        return '{cte}."{pk}" = {table}."{pk}"'.format(
            cte=query.model._cte_node_table,
            pk=query.model._meta.pk.attname,
            table=maybe_alias(query.model._meta.db_table))

    @classmethod
    def _remove_cte_where(cls, query):
        where = cls._generate_where(query)
        for w in query.where.children:
            if isinstance(w, ExtraWhere):
                if where in w.sqls:
                    query.where.children.remove(w)

    def get_compiler(self, using=None, connection=None):
        """ Overrides the Query method get_compiler in order to return
            an instance of the above custom compiler.
        """
        # Copy the body of this method from Django except the final
        # return statement. We will ignore code coverage for this.
        if using is None and connection is None:  # pragma: no cover
            raise ValueError("Need either using or connection")
        if using:
            connection = connections[using]
        # Check that the compiler will be able to execute the query
        for alias, aggregate in self.annotation_select.items():
            connection.ops.check_expression_support(aggregate)
        # Instantiate the custom compiler.
        return {
            CTEUpdateQuery: CTEUpdateQueryCompiler,
            CTEInsertQuery: CTEInsertQueryCompiler,
            CTEDeleteQuery: CTEDeleteQueryCompiler,
            CTEAggregateQuery: CTEAggregateQueryCompiler,
        }.get(self.__class__, CTEQueryCompiler)(self, connection, using)

    if django.VERSION < (2, 0):
        def clone(self, klass=None, memo=None, **kwargs):
            """ Overrides Django's Query clone in order to return appropriate CTE
                compiler based on the target Query class. This mechanism is used by
                methods such as 'update' and '_update' in order to generate UPDATE
                queries rather than SELECT queries.
            """
            klass = {
                UpdateQuery: CTEUpdateQuery,
                InsertQuery: CTEInsertQuery,
                DeleteQuery: CTEDeleteQuery,
                AggregateQuery: CTEAggregateQuery,
            }.get(klass, self.__class__)
            return super(CTEQuery, self).clone(klass, memo, **kwargs)

    else:
        def chain(self, klass=None):
            """ Overrides Django's Query clone in order to return appropriate CTE
                compiler based on the target Query class. This mechanism is used by
                methods such as 'update' and '_update' in order to generate UPDATE
                queries rather than SELECT queries.
            """
            klass = {
                UpdateQuery: CTEUpdateQuery,
                InsertQuery: CTEInsertQuery,
                DeleteQuery: CTEDeleteQuery,
                AggregateQuery: CTEAggregateQuery,
            }.get(klass, self.__class__)
            return super(CTEQuery, self).chain(klass)


class CTEUpdateQuery(UpdateQuery, CTEQuery):
    pass


class CTEInsertQuery(InsertQuery, CTEQuery):
    pass


class CTEDeleteQuery(DeleteQuery, CTEQuery):
    pass


class CTEAggregateQuery(AggregateQuery, CTEQuery):
    pass


class CTECompiler(object):

    CTE = """WITH RECURSIVE {cte} ("{depth}", "{path}", "{ordering}", "{pk}") AS (

    SELECT 1 AS depth,
           array[{pk_path}] AS {path},
           {order} AS {ordering},
           T."{pk}"
      FROM {db_table} T
     WHERE T."{parent}" IS NULL

     UNION ALL

    SELECT {cte}.{depth} + 1 AS {depth},
           {cte}.{path} || {pk_path},
           {cte}.{ordering} || {order},
           T."{pk}"
      FROM {db_table} T
      JOIN {cte} ON T."{parent}" = {cte}."{pk}")

    """

    @classmethod
    def generate_sql(cls, connection, query, as_sql):
        """

        :param query:
        :type query:
        :param sql:
        :type sql:
        :return:
        :rtype:
        """
        # add the CTE table to the Query's extras precisely once (because we
        # could be combining multiple CTE Queries).
        if query.model._cte_node_table not in query.extra_tables:
            query.add_extra(
                select=None,
                select_params=None,
                where=None,
                params=None,
                tables=[query.model._cte_node_table],
                order_by=None)

        # place appropriate CTE table prefix to any order_by or extra_order_by
        # entries which reference virtual fields, and preserve the optional
        # sign.
        cte_columns = (
            query.model._cte_node_depth,
            query.model._cte_node_path,
            query.model._cte_node_ordering,
        )

        def maybe_prefix_order_by(order_by):
            return [
                '{sign}{cte}.{column}'.format(
                    sign='-' if order[0] == '-' else '',
                    cte=query.model._cte_node_table,
                    column=order.lstrip('-'))
                if order.lstrip('-') in cte_columns
                else order
                for order in order_by
            ]
        query.order_by = maybe_prefix_order_by(query.order_by)
        query.extra_order_by = maybe_prefix_order_by(query.extra_order_by)

        # Obtain compiled SQL from Django.
        sql = as_sql()

        def maybe_cast(field):
            # If the ordering information specified a type to cast to, then use
            # this type immediately, otherwise determine whether a
            # variable-length character field should be cast into TEXT or if no
            # casting is necessary. A None type defaults to the latter.
            if type(field) == tuple and not field[1] is None:
                return 'CAST (T."%s" AS %s)' % field
            else:
                if type(field) == tuple:
                    name = field[0]
                else:
                    name = field
                _field = query.model._meta.get_field(name)
                if _field.db_type(connection).startswith('varchar'):
                    return 'CAST (T."%s" AS TEXT)' % name
                else:
                    return 'T."%s"' % name

        # The primary key is used in the path; in case it is of a custom type,
        # ensure appropriate casting is performed. This is a very rare
        # condition, as most types can be used directly in the path, especially
        # since no other fields with incompatible types are combined (with a
        # notable exception of VARCHAR types which must be converted to TEXT).
        pk_path = maybe_cast((query.model._meta.pk.attname,
                              query.model._cte_node_primary_key_type))

        # If no explicit ordering is specified, then use the primary key. If the
        # primary key is used in ordering, and it is of a type which needs
        # casting in order to be used in the ordering, then it is possible that
        # explicit casting was not specified through _cte_node_order_by because
        # it is expected to be specified through the _cte_node_primary_key_type
        # attribute. Specifying the cast type of the primary key in the
        # _cte_node_order_by attribute has precedence over
        # _cte_node_primary_key_type.
        if not hasattr(query.model, '_cte_node_order_by') or \
                query.model._cte_node_order_by is None or \
                len(query.model._cte_node_order_by) == 0:
            order = 'array[%s]' % maybe_cast((
                query.model._meta.pk.attname,
                query.model._cte_node_primary_key_type))
        else:
            # Compute the ordering virtual field constructor, possibly casting
            # fields into a common type.
            order = '||'.join(
                'array[%s]' % maybe_cast(field)
                for field in query.model._cte_node_order_by
            )
        # Prepend the CTE with the ordering constructor and table
        # name to the SQL obtained from Django above.
        return (''.join([
            cls.CTE.format(
                order=order,
                cte=query.model._cte_node_table,
                depth=query.model._cte_node_depth,
                path=query.model._cte_node_path,
                ordering=query.model._cte_node_ordering,
                parent=query.model._cte_node_parent_attname,
                pk=query.model._meta.pk.attname,
                pk_path=pk_path,
                db_table=query.model._meta.db_table
            ),
            sql[0],
        ]), sql[1])


class CTEQueryCompiler(SQLCompiler):

    def as_sql(self, *args, **kwargs):
        """
        Overrides the :class:`SQLCompiler` method in order to prepend the
        necessary CTE syntax, as well as perform pre- and post-processing,
        including adding the extra CTE table and WHERE clauses.

        :param with_limits:
        :type with_limits:
        :param with_col_aliases:
        :type with_col_aliases:
        :return:
        :rtype:
        """
        def _as_sql():
            return super(CTEQueryCompiler, self).as_sql(*args, **kwargs)
        return CTECompiler.generate_sql(self.connection, self.query, _as_sql)


class CTEUpdateQueryCompiler(SQLUpdateCompiler):

    def as_sql(self, *args, **kwargs):
        """
        Overrides the :class:`SQLUpdateCompiler` method in order to remove any
        CTE-related WHERE clauses, which are not necessary for UPDATE queries,
        yet may have been added if this query was cloned from a CTEQuery.

        :return:
        :rtype:
        """
        CTEQuery._remove_cte_where(self.query)
        return super(self.__class__, self).as_sql(*args, **kwargs)


class CTEInsertQueryCompiler(SQLInsertCompiler):

    def as_sql(self, *args, **kwargs):
        """
        Overrides the :class:`SQLInsertCompiler` method in order to remove any
        CTE-related WHERE clauses, which are not necessary for INSERT queries,
        yet may have been added if this query was cloned from a CTEQuery.

        :return:
        :rtype:
        """
        CTEQuery._remove_cte_where(self.query)
        return super(self.__class__, self).as_sql(*args, **kwargs)


class CTEDeleteQueryCompiler(SQLDeleteCompiler):

    def as_sql(self, *args, **kwargs):
        """
        Overrides the :class:`SQLDeleteCompiler` method in order to remove any
        CTE-related WHERE clauses, which are not necessary for DELETE queries,
        yet may have been added if this query was cloned from a CTEQuery.

        :return:
        :rtype:
        """
        CTEQuery._remove_cte_where(self.query)
        return super(self.__class__, self).as_sql(*args, **kwargs)


class CTEAggregateQueryCompiler(SQLAggregateCompiler):

    def as_sql(self, *args, **kwargs):
        """
        Overrides the :class:`SQLAggregateCompiler` method in order to
        prepend the necessary CTE syntax, as well as perform pre- and post-
        processing, including adding the extra CTE table and WHERE clauses.

        :param qn:
        :type qn:
        :return:
        :rtype:
        """
        def _as_sql():
            return super(CTEAggregateQueryCompiler, self).as_sql(*args, **kwargs)
        return CTECompiler.generate_sql(self.connection, self.query, _as_sql)



# -*- coding: utf-8 -*-
#
# This document is free and open-source software, subject to the OSI-approved
# BSD license below.
#
# Copyright (c) 2011 - 2013 Alexis Petrounias <www.petrounias.org>,
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the author nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

""" Django CTE Trees Models.
"""
from __future__ import unicode_literals

from django.core.exceptions import (
    FieldError, ImproperlyConfigured, ValidationError,
)
from django.db.models import CASCADE, ForeignKey, Manager, Model
from django.db.models.fields import FieldDoesNotExist
from django.utils.translation import ugettext as _

from .query import CTEQuerySet


class CTENodeManager(Manager):
    """ Custom :class:`Manager` which ensures all queries involving
        :class:`CTENode` objects are processed by the custom SQL compiler.
        Additionally, provides tree traversal queries for obtaining node
        children, siblings, ancestors, descendants, and roots.

        If your Model inherits from :class:`CTENode` and use your own custom
        :class:`Manager`, you must ensure the following three:

        1) your :class:`Manager` inherits from :class:`CTENodeManager`,

        2) if you override the :meth:`get_queryset` method in order to
        return a custom :class:`QuerySet`, then your `QuerySet` must also
        inherit from :class:`CTENodeManager.CTEQuerySet`, and

        3) invoke the :meth:`_ensure_parameters` on your :class:`Manager`
        at least once before using a :class:`QuerySet` which inherits from
        :class:`CTENodeManager.CTEQuerySet`, unless you have supplied the
        necessary CTE node attributes on the :class:`CTENode` :class:`Model` in
        some other way.

        The methods :meth:`prepare_delete`, :meth:`prepare_delete_pharaoh`,
        :meth:`prepare_delete_grandmother`, and
        :meth:`prepare_delete_monarchy` can be directly used to prepare
        nodes for deletion with either the default or explicitly-specified
        deletion semantics. The :class:`CTENode` abstract :class:`Model`
        defines a :meth:`CTENode.delete` method which delegates preparation
        to this manager.

    """
    # SQL CTE temporary table name.
    DEFAULT_TABLE_NAME = 'cte'
    DEFAULT_CHILDREN_NAME = 'children'

    # Tree traversal semantics.
    TREE_TRAVERSAL_NONE = 'none'
    TREE_TRAVERSAL_DFS = 'dfs'
    TREE_TRAVERSAL_BFS = 'bfs'
    TREE_TRAVERSAL_METHODS = (TREE_TRAVERSAL_NONE, TREE_TRAVERSAL_DFS,
                              TREE_TRAVERSAL_BFS)
    TREE_TRAVERSAL_CHOICES = (
        (TREE_TRAVERSAL_NONE, _('none')),
        (TREE_TRAVERSAL_DFS, _('depth first search')),
        (TREE_TRAVERSAL_BFS, _('breadth first search')),
    )
    DEFAULT_TREE_TRAVERSAL = TREE_TRAVERSAL_DFS

    # Virtual fields.
    VIRTUAL_FIELD_DEPTH = 'depth'
    VIRTUAL_FIELD_PATH = 'path'
    VIRTUAL_FIELD_ORDERING = 'ordering'

    # Deletion semantics.
    DELETE_METHOD_NONE = 'none'
    DELETE_METHOD_PHARAOH = 'pharaoh'
    DELETE_METHOD_GRANDMOTHER = 'grandmother'
    DELETE_METHOD_MONARCHY = 'monarchy'
    DELETE_METHODS = (DELETE_METHOD_NONE, DELETE_METHOD_PHARAOH,
                      DELETE_METHOD_GRANDMOTHER, DELETE_METHOD_MONARCHY)
    DELETE_METHOD_CHOICES = (
        (DELETE_METHOD_NONE, _('none')),
        (DELETE_METHOD_PHARAOH, _('pharaoh (all subtree)')),
        (DELETE_METHOD_GRANDMOTHER, _('grandmother (move subtree up)')),
        (DELETE_METHOD_MONARCHY,
         _('monarchy (first child becomes subtree root)')),
    )
    DEFAULT_DELETE_METHOD = DELETE_METHOD_PHARAOH

    # Related manager lookup should return this custom Manager in order to use
    # the custom QuerySet above.
    use_for_related_fields = True
    silence_use_for_related_fields_deprecation = True

    def _ensure_parameters(self):
        """ Attempts to load and verify the CTE node parameters. Will use
            default values for all missing parameters, and raise an exception if
            a parameter's value cannot be verified. This method will only
            perform these actions once, and set the :attr:`_parameters_checked`
            attribute to ``True`` upon its first success.
        """

        if hasattr(self, '_parameters_checked'):
            return

        if not hasattr(self.model, '_cte_node_table') or \
                self.model._cte_node_table is None:
            setattr(self.model, '_cte_node_table',
                    self.DEFAULT_TABLE_NAME)

        if not hasattr(self.model, '_cte_node_depth') or \
                self.model._cte_node_depth is None:
            setattr(self.model, '_cte_node_depth',
                    self.VIRTUAL_FIELD_DEPTH)

        if not hasattr(self.model, '_cte_node_path') or \
                self.model._cte_node_depth is None:
            setattr(self.model, '_cte_node_path',
                    self.VIRTUAL_FIELD_PATH)

        if not hasattr(self.model, '_cte_node_ordering') or \
                self.model._cte_node_ordering is None:
            setattr(self.model, '_cte_node_ordering',
                    self.VIRTUAL_FIELD_ORDERING)

        if not hasattr(self.model, '_cte_node_traversal') or \
                self.model._cte_node_traversal is None:
            setattr(self.model, '_cte_node_traversal',
                    self.DEFAULT_TREE_TRAVERSAL)

        if not hasattr(self.model, '_cte_node_children') or \
                self.model._cte_node_children is None:
            setattr(self.model, '_cte_node_children',
                    self.DEFAULT_CHILDREN_NAME)

        if not hasattr(self.model, '_cte_node_primary_key_type'):
            setattr(self.model, '_cte_node_primary_key_type', None)

        # Determine the parent foreign key field name, either
        # explicitly specified, or the first foreign key to 'self'.
        # If we need to determine, then we set the attribute for future
        # reference.
        if not hasattr(self.model, '_cte_node_parent') or \
                self.model._cte_node_parent is None:
            found = False
            for f in self.model._meta.fields:
                if isinstance(f, ForeignKey):
                    if f.remote_field.model == self.model:
                        setattr(self.model, '_cte_node_parent', f.name)
                        found = True
            if not found:
                raise ImproperlyConfigured(
                    _('CTENode must have a Foreign Key to self for the parent '
                      'relation.'))

        try:
            parent_field = self.model._meta.get_field(
                self.model._cte_node_parent)
        except FieldDoesNotExist:
            raise ImproperlyConfigured(''.join([
                _('CTENode._cte_node_parent must specify a Foreign Key to self, '
                  'instead it is: '),
                self.model._cte_node_parent]))

        # Ensure parent relation is a Foreign Key to self.
        if not parent_field.remote_field.model == self.model:
            raise ImproperlyConfigured(''.join([
                _('CTENode._cte_node_parent must specify a Foreign Key to self, '
                  'instead it is: '),
                self.model._cte_node_parent]))

        # Record the parent field attribute name for future reference.
        setattr(
            self.model, '_cte_node_parent_attname',
            self.model._meta.get_field(self.model._cte_node_parent).attname)

        # Ensure traversal choice is valid.
        traversal_choices = [
            choice[0] for choice in self.TREE_TRAVERSAL_CHOICES]
        if self.model._cte_node_traversal not in traversal_choices:
            raise ImproperlyConfigured(
                ' '.join(['CTENode._cte_node_traversal must be one of [',
                          ', '.join(traversal_choices), ']; instead it is:',
                          self.model._cte_node_traversal]))

        # Ensure delete choice is valid.
        if not hasattr(self.model, '_cte_node_delete_method') or \
                self.model._cte_node_delete_method is None:
            setattr(self.model, '_cte_node_delete_method',
                    self.DEFAULT_DELETE_METHOD)
        else:
            # Ensure specified method is valid.
            method_choices = [
                dm[0] for dm in self.DELETE_METHOD_CHOICES]
            if self.model._cte_node_delete_method not in method_choices:
                raise ImproperlyConfigured(
                    ' '.join(['delete method must be one of [',
                              ', '.join(method_choices), ']; instead it is:',
                              self.model._cte_node_delete_method]))

        setattr(self, '_parameters_checked', True)

    def _ensure_virtual_fields(self, node):
        """ Attempts to read the virtual fields from the given `node` in order
            to ensure they exist, resulting in an early :class:`AttributeError`
            exception in case of missing virtual fields. This method requires
            several parameters, and thus invoked :meth:`_ensure_parameters`
            first, possibly resulting in an :class:`ImproperlyConfigured`
            exception being raised.

            :param node: the :class:`CTENode` for which to verify that the
                virtual fields are present.
        """
        # Uses several _cte_node_* parameters, so ensure they exist.
        self._ensure_parameters()
        for vf in [
                self.model._cte_node_depth,
                self.model._cte_node_path,
                self.model._cte_node_ordering]:
            if not hasattr(node, vf):
                raise FieldError(
                    _('CTENode objects must be loaded from the database before '
                      'they can be used.'))

    def get_queryset(self):
        """ Returns a custom :class:`QuerySet` which provides the CTE
            functionality for all queries concerning :class:`CTENode` objects.
            This method overrides the default :meth:`get_queryset` method of
            the :class:`Manager` class.

            :returns: a custom :class:`QuerySet` which provides the CTE
                functionality for all queries concerning :class:`CTENode`
                objects.
        """
        # The CTEQuerySet uses _cte_node_* attributes from the Model, so ensure
        # they exist.
        self._ensure_parameters()
        return CTEQuerySet(self.model, using=self._db)

    def roots(self):
        """ Returns a :class:`QuerySet` of all root :class:`CTENode` objects.

            :returns: a :class:`QuerySet` of all root :class:`CTENode` objects.
        """
        # We need to read the _cte_node_parent attribute, so ensure it exists.
        self._ensure_parameters()
        # We need to construct: self.filter(parent = None)
        return self.filter(**{self.model._cte_node_parent: None})

    def leaves(self):
        """ Returns a :class:`QuerySet` of all leaf nodes (nodes with no
            children).

            :return: A :class:`QuerySet` of all leaf nodes (nodes with no
                children).
        """
        # We need to read the _cte_node_children attribute, so ensure it exists.
        self._ensure_parameters()
        return self.exclude(**{
            '%s__id__in' % self.model._cte_node_children: self.all(),
        })

    def branches(self):
        """ Returns a :class:`QuerySet` of all branch nodes (nodes with at least
            one child).

            :return: A :class:`QuerySet` of all leaf nodes (nodes with at least
                one child).
        """
        # We need to read the _cte_node_children attribute, so ensure it exists.
        self._ensure_parameters()
        return self.filter(**{
            '%s__id__in' % self.model._cte_node_children: self.all(),
        }).distinct()

    def root(self, node):
        """ Returns the :class:`CTENode` which is the root of the tree in which
            the given `node` participates (or `node` if it is a root node). This
            method functions through the :meth:`get` method.

            :param node: A :class:`CTENode` whose root is required.

            :return: A :class:`CTENode` which is the root of the tree in which
                the given `node` participates (or the given `node` if it is a
                root node).
        """
        # We need to use the path virtual field, so ensure it exists.
        self._ensure_virtual_fields(node)
        return self.get(pk=getattr(node, self.model._cte_node_path)[0])

    def siblings(self, node):
        """ Returns a :class:`QuerySet` of all siblings of a given
            :class:`CTENode` `node`.

            :param node: a :class:`CTENode` whose siblings are required.

            :returns: A :class:`QuerySet` of all siblings of the given `node`.
        """
        # We need to read the _cte_node_parent* attributes, so ensure they
        # exist.
        self._ensure_parameters()
        # We need to construct: filter(parent = node.parent_id)
        return self.filter(**{
            self.model._cte_node_parent: getattr(
                node,
                self.model._cte_node_parent_attname,
            ),
        }).exclude(id=node.id)

    def ancestors(self, node):
        """ Returns a :class:`QuerySet` of all ancestors of a given
            :class:`CTENode` `node`.

            :param node: A :class:`CTENode` whose ancestors are required.

            :returns: A :class:`QuerySet` of all ancestors of the given `node`.
        """
        # We need to use the path virtual field, so ensure it exists.
        self._ensure_virtual_fields(node)
        return self.filter(
            pk__in=getattr(node, self.model._cte_node_path)[:-1])

    def descendants(self, node):
        """ Returns a :class:`QuerySet` with all descendants for a given
            :class:`CTENode` `node`.

            :param node: the :class:`CTENode` whose descendants are required.

            :returns: A :class:`QuerySet` with all descendants of the given
                `node`.
        """
        # We need to read the _cte_node_* attributes, so ensure they exist.
        self._ensure_parameters()
        # This is implemented in the CTE WHERE logic, so we pass a reference to
        # the offset CTENode to the custom QuerySet, which will process it.
        # Because the compiler will include the node in question in the offset,
        # we must exclude it here.
        return CTEQuerySet(self.model, using=self._db,
                           offset=node).exclude(pk=node.pk)

    def is_parent_of(self, node, subject):
        """ Returns ``True`` if the given `node` is the parent of the given
            `subject` node, ``False`` otherwise. This method uses the
            :attr:`parent` field, and so does not perform any query.

            :param node: the :class:`CTENode' for which to determine whether it
                is a parent of the `subject`.

            :param subject: the :class:`CTENode` for which to determine whether
                its parent is the `node`.

            :returns: ``True`` if `node` is the parent of `subject`, ``False``
                otherwise.
        """
        return subject.parent_id == node.id

    def is_child_of(self, node, subject):
        """ Returns ``True`` if the given `node` is a child of the given
            `subject` node, ``False`` otherwise. This method used the
            :attr:`parent` field, and so does not perform any query.

            :param node: the :class:`CTENode' for which to determine whether it
                is a child of the `subject`.

            :param subject: the :class:`CTENode` for which to determine whether
                one of its children is the `node`.

            :returns: ``True`` if `node` is a child of `subject`, ``False``
                otherwise.
        """
        return node.parent_id == subject.id

    def is_sibling_of(self, node, subject):
        """ Returns ``True`` if the given `node` is a sibling of the given
            `subject` node, ``False`` otherwise. This method uses the
            :attr:`parent` field, and so does not perform any query.

            :param node: the :class:`CTENode' for which to determine whether it
                is a sibling of the `subject`.

            :param subject: the :class:`CTENode` for which to determine whether
                one of its siblings is the `node`.

            :returns: ``True`` if `node` is a sibling of `subject`, ``False``
                otherwise.
        """
        # Ensure nodes are not siblings of themselves.
        return not node.id == subject.id and node.parent_id == subject.parent_id

    def is_ancestor_of(self, node, subject):
        """ Returns ``True`` if the given `node` is an ancestor of the given
            `subject` node, ``False`` otherwise. This method uses the
            :attr:`path` virtual field, and so does not perform any query.

            :param node: the :class:`CTENode' for which to determine whether it
                is an ancestor of the `subject`.

            :param subject: the :class:`CTENode` for which to determine whether
                one of its ancestors is the `node`.

            :returns: ``True`` if `node` is an ancestor of `subject`, ``False``
                otherwise.
        """
        # We will need to use the path virtual field, so ensure it exists.
        self._ensure_virtual_fields(node)

        # Convenience check so is_ancestor_of can be combined with methods
        # returning nodes without the caller having to worry about a None
        # subject.
        if subject is None:
            return False

        # Optimization: a node will never be an ancestor of a root node.
        if getattr(subject, subject._cte_node_depth) == 1:
            return False

        # The path will either be an index of primitives, or an encoding of an
        # array.
        if type(getattr(node, node._cte_node_path)) == list:
            # We can slice with -1 because we know that depth > 1 from above.
            return node.id in getattr(subject, subject._cte_node_path)[0:-1]
        else:
            # Search for node id up to the penultimate entry in the path of the
            # subject, meaning we ignore the end of the path consisting of:
            # a) the ending closing curly brace,
            # b) the length of the subject id, and
            # c) the separator character,
            # therefore we look for a match ending at the length of the
            # subject's id string plus two (so negative index length minus two).
            _path = getattr(subject, subject._cte_node_path)
            return _path[:-len(str(subject.id)) - 2].index(str(node.id)) > 0

    def is_descendant_of(self, node, subject):
        """ Returns ``True`` if the given `node` is a descendant of the given
            `subject` node, ``False`` otherwise. This method uses the
            :attr:`path` virtual field, and so does not perform any query.

            :param node: the :class:`CTENode' for which to determine whether it
                is a descendant of the `subject`.

            :param subject: the :class:`CTENode` for which to determine whether
                one of its descendants is the `node`.

            :returns: ``True`` if `node` is a descendant of `subject`, ``False``
                otherwise.
        """
        # We will need to use the path virtual field, so ensure it exists.
        self._ensure_virtual_fields(node)

        # Convenience check so is_descendant_of can be combined with methods
        # returning nodes without the caller having to worry about a None
        # subject.
        if subject is None:
            return False

        # Optimization: a root node will never be a descendant of any node.
        if getattr(node, node._cte_node_depth) == 1:
            return False

        # The path will either be an index of primitives, or an encoding of an
        # array.
        if type(getattr(node, node._cte_node_path)) == list:
            # We can slice with -1 because we know that depth > 1 from above.
            return subject.id in getattr(node, node._cte_node_path)[0:-1]
        else:
            # Search for subject id up to the penultimate entry in the path of
            # the node, meaning we ignore the end of the path consisting of:
            # a) the ending closing curly brace,
            # b) the length of the node id, and
            # c) the separator character,
            # therefore we look for a match ending at most at the length of the
            # node's id string plus two (so negative index length minus two).
            _path = getattr(node, node._cte_node_path)
            return _path[:-len(str(node.id)) - 2].index(str(subject.id)) > 0

    def is_leaf(self, node):
        """ Returns ``True`` if the given `node` is a leaf (has no children),
            ``False`` otherwise.

            :param node: the :class:`CTENode` for which to determine whether it
                is a leaf.

            :return: ``True`` if `node` is a leaf, ``False`` otherwise.
        """
        return not node.children.exists()

    def is_branch(self, node):
        """ Returns ``True`` if the given `node` is a branch (has at least one
            child), ``False`` otherwise.

            :param node: the :class:`CTENode` for which to determine whether it
                is a branch.

            :return: ``True`` if `node` is a branch, ``False`` otherwise.
        """
        return node.children.exists()

    def attribute_path(
        self, node, attribute, missing=None,
        visitor=lambda node, attribute: getattr(node, attribute, None),
    ):
        """ Generates a list of values of the `attribute` of all ancestors of
            the given `node` (as well as the node itself). If a value is
            ``None``, then the optional value of `missing` is used (by default
            ``None``).

            By default, the ``getattr(node, attribute, None) or missing``
            mechanism is used to obtain the value of the attribute for each
            node. This can be overridden by supplying a custom `visitor`
            function, which expects as arguments the node and the attribute, and
            should return an appropriate value for the required attribute.

            :param node: the :class:`CTENode` for which to generate the
                attribute path.

            :param attribute: the name of the attribute.

            :param missing: optional value to use when attribute value is None.

            :param visitor: optional function responsible for obtaining the
                attribute value from a node.

            :return: a list of values of the required `attribute` of the
                ancestor path of the given `node`.
        """
        return [
            visitor(c, attribute) or missing
            for c in node.ancestors()
        ] + [
            visitor(node, attribute) or missing
        ]

    def as_tree(self, visitor=None, children=None):
        """ Recursively traverses each tree (starting from each root) in order
            to generate a dictionary-based tree structure of the entire forest.
            Each level of the forest/tree is a list of nodes, and each node
            consists of a dictionary representation, where the entry
            ``children`` (by default) consists of a list of dictionary
            representations of its children.

            Optionally, a `visitor` callback can be used, which is responsible
            for generating a dictionary representation of a given
            :class:`CTENode`. By default, the :meth:`_default_node_visitor` is
            used which generates a dictionary with the current node as well as
            structural properties. See :meth:`_default_node_visitor` for the
            appropriate signature of this callback.

            Optionally, a `children` callback can be used, which is responsible
            for determining which :class:`CTENode`s are children of each visited
            :class:`CTENode`, resulting in a key (by default ``children``) and a
            list of children :class:`CTENode` objects, which are then included
            in the dictionary representation of the currently-visited node. See
            :meth:`_default_node_children` for the appropriate signature of this
            callback.

            For each node visited, the :meth:`CTENode.as_tree` method is invoked
            along with the optional `visitor` and `children` arguments. This
            method, if not overridden, will delegate to :meth:`node_as_tree`,
            which is responsible for invoking the :meth:`visitor` and
            :meth:`children` methods, as well as updating the dictionary
            representation of the node with the representation of the children
            nodes.

            :param visitor: optional function responsible for generating the
                dictionary representation of a node.

            :param children: optional function responsible for generating a
                children key and list for a node.

            :return: a dictionary representation of the structure of the forest.
        """
        return [
            root.as_tree(visitor=visitor, children=children)
            for root in self.roots()
        ]

    def node_as_tree(
        self,
        node,
        visitor=lambda self, node: self._default_node_visitor(node),
        children=lambda self, node, visitor, children:
            self._default_node_children(node, visitor, children),
    ):
        """ Visits a :class:`CTENode` `node` and delegates to the (optional)
            `visitor` callback, as well as the (optional) `children` callback,
            in order to generate a dictionary representation of the node along
            with its children nodes.

            :param node: the :class:`CTENode` for which to generate the
                representation.

            :param visitor: optional function responsible for generating the
                dictionary representation of the node.

            :param children: optional function responsible for generating a
                children key and list for the node.

            :return: a dictionary representation of the structure of the node
                and its descendant tree.
        """
        tree = visitor(self, node)
        tree.update(children(self, node, visitor, children))
        return tree

    def _default_node_visitor(self, node):
        """ Generates a dictionary representation of the given :class:`CTENode`
        `node`, which consists of the node itself under the key ``node``, as
        well as structural information under the keys ``depth``, ``path``,
        ``ordering``, ``leaf``, and ``branch``.

        :param node: the :class:`CTENode` for which to generate the
            representation.

        :return: a dictionary representation of the structure of the node.
        """
        return {
            'depth': getattr(node, node._cte_node_depth),
            'path': [str(c) for c in getattr(node, node._cte_node_path)],
            'ordering': getattr(node, node._cte_node_ordering),
            'leaf': node.is_leaf(),
            'branch': node.is_branch(),
            'node': node,
        }

    def _default_node_children(self, node, visitor, children):
        """ Generates a key and list of children of the given :class:`CTENode`
        `node`, intended to be used as an update to the dictionary
        representation generated by the :meth:`node_as_tree` method. The key is
        ``children`` and the list consists of the children of the given node as
        determined by the `children` callback.

        Each child node is, in turn, visited through recursive calls to
        :meth:`node_as_child`, and the `visitor` and `children` parameters are
        passed along.

        :param node: the :class:`CTENode` for which to generate the children
            representation.

        :param visitor: optional function responsible for generating the
            dictionary representation of the node.

        :param children: optional function responsible for generating a
            children key and list for the node.

        :return: a key and list representation of the structure of the children
            of the given node.
        """
        return {
            self.model._cte_node_children: [
                self.node_as_tree(
                    child,
                    visitor=visitor,
                    children=children,
                ) for child in node.children.all()
            ],
        }

    def drilldown(self, attributes, path):
        """ Recursively descends the tree/forest (starting from each root node)
            in order to find a :class:`CTENode` which corresponds to the given
            `path`. The path is expected to be an iterable of tuples, called
            path components, consisting of attribute values with which to filter
            through the QuerySet API. The name of the attribute to which each
            value corresponds is specified in `attributes`, which is expected
            to conform to Django's QuerySet API for the filter semantics. Each
            value in the path component tuple will be mapped to its
            corresponding attribute name before being passed to the filter
            method.

            For example, if the node model features the integer field ``x``
            and the boolean field ``y``, we can drill down in the following way:

                drilldown(('x__gte', 'y'),[(35, True), (37, False)])

            The simplest form of drilldown is to match with equality on a single
            attribute, such as ``name``, as in the following example:

                drilldown(('name',), [('path',), ('to',), ('my',), ('node',)])

            Don't forget the trailing comma if specifying singleton tuples! If
            you need very simple, one-attribute path components, it is suggested
            you extend the manager with your own convenience method; the above
            will, for instance, become:

                def drilldown_by_name(self, path):
                    return self.drilldown(('name',),
                        [(component,) for component in path])

            Failure to find the required node results in a :class:`DoesNotExist`
            exception being raised.

            An empty path will result in the first root node being returned (if
            at least one root node exists).
        """

        # empty path result in first root, if present
        if len(path) == 0:
            try:
                return self.roots()[0]
            except IndexError:
                raise self.model.DoesNotExist

        # bootstrap with the first component, then iterate through the remaining
        # components in the path as long as each child node is found
        component = path[0]
        current = None

        # mapping of attribute names with values, as per QuerySet filter
        def attrs(component):
            return dict(zip(attributes, component))

        # find the root corresponding to the bootstrapped initial path component
        try:
            root = self.roots().filter(**attrs(component))[0]
        except IndexError:
            raise self.model.DoesNotExist

        # proceed to drill down until path components are exhausted
        current = root
        for component in path[1:]:
            try:
                current = current.children.filter(**attrs(component))[0]
            except IndexError:
                raise self.model.DoesNotExist
        return current

    def prepare_delete(self, node, method, position=None, save=True):
        """ Prepares a given :class:`CTENode` `node` for deletion, by executing
            the required deletion semantics (Pharaoh, Grandmother, or Monarchy).

            The `method` argument can be one of the valid
            :const:`DELETE_METHODS` choices. If it is
            :const:`DELETE_METHOD_NONE` or ``None``, then the default delete
            method will be used (as specified from the optional
            :attr:`_cte_node_delete_method`).

            Under the :const:`DELETE_METHOD_GRANDMOTHER` and
            :const:`DELETE_METHOD_MONARCHY` delete semantics, descendant nodes
            may be moved; in this case the optional `position` can be a
            ``callable`` which is invoked prior to each move operation (see
            :meth:`move` for details).

            Furthermore, by default, after each move operation, sub-tree nodes
            which were moved will be saved through a call to :meth:`Model.save`
            unless `save` is ``False``.

            This method delegates move operations to :meth:`move`.

            :param node: the :class:`CTENode` to prepare for deletion.

            :param method: optionally, a delete method to use.

            :param position: optionally, a ``callable`` to invoke prior to each
                move operation.

            :param save: flag indicating whether to save after each move
                operation, ``True`` by default.
        """
        # If no delete method preference is specified, use attribute.
        if method is None:
            method = node._cte_node_delete_method
        # If no preference specified, use default.
        if method == self.DELETE_METHOD_NONE:
            method = self.DEFAULT_DELETE_METHOD
        # Delegate to appropriate method.
        getattr(self, 'prepare_delete_%s' % method)(node, position, save)

    def prepare_delete_pharaoh(self, node, position=None, save=True):
        """ Prepares a given :class:`CTENode` `node` for deletion, by executing
            the :const:`DELETE_METHOD_PHARAOH` semantics.

            This method does not perform any sub-tree reorganization, and hence
            no move operation, so the `position` and `save` arguments are
            ignored; they are present for regularity purposes with the rest of
            the deletion preparation methods.

            :param node: the :class:`CTENode` to prepare for deletion.

            :param position: this is ignored, but present for regularity.

            :param save: this is ignored, but present for regularity.
        """
        # Django will take care of deleting the sub-tree through the reverse
        # Foreign Key parent relation.
        pass

    def prepare_delete_grandmother(self, node, position=None, save=True):
        """ Prepares a given :class:`CTENode` `node` for deletion, by executing
            the :const:`DELETE_METHOD_GRANDMOTHER` semantics. Descendant nodes,
            if present, will be moved; in this case the optional `position` can
            be a ``callable`` which is invoked prior to each move operation (see
            :meth:`move` for details).

            By default, after each move operation, sub-tree nodes which were
            moved will be saved through a call to :meth:`Model.save` unless
            `save` is ``False``.

            This method delegates move operations to :meth:`move`.

            :param node: the :class:`CTENode` to prepare for deletion.

            :param position: optionally, a ``callable`` to invoke prior to each
                move operation.

            :param save: flag indicating whether to save after each move
                operation, ``True`` by default.
        """
        # Move all children to the node's parent.
        for child in node.children.all():
            child.move(node.parent, position, save)

    def prepare_delete_monarchy(self, node, position=None, save=True):
        """ Prepares a given :class:`CTENode` `node` for deletion, by executing
            the :const:`DELETE_METHOD_MONARCHY` semantics. Descendant nodes,
            if present, will be moved; in this case the optional `position` can
            be a ``callable`` which is invoked prior to each move operation (see
            :meth:`move` for details).

            By default, after each move operation, sub-tree nodes which were
            moved will be saved through a call to :meth:`Model.save` unless
            `save` is ``False``.

            This method delegates move operations to :meth:`move`.

            :param node: the :class:`CTENode` to prepare for deletion.

            :param position: optionally, a ``callable`` to invoke prior to each
                move operation.

            :param save: flag indicating whether to save after each move
                operation, ``True`` by default.
        """
        # We are going to iterate all children, even though the first child is
        # treated in a special way, because the query iterator may be custom, so
        # we will avoid using slicing children[0] and children[1:].
        first = None
        for child in node.children.all():
            if first is None:
                first = child
                first.move(node.parent, position, save)
            else:
                child.move(first, position, save)

    def move(self, node, destination, position=None, save=False):
        """ Moves the given :class:`CTENode` `node` and places it as a child
            node of the `destination` :class:`CTENode` (or makes it a root node
            if `destination` is ``None``).

            Optionally, `position` can be a callable which is invoked prior to
            placement of the `node` with the `node` and the `destination` node
            as the sole two arguments; this can be useful in implementing
            specific sibling ordering semantics.

            Optionally, if `save` is ``True``, after the move operation
            completes (after the :attr:`CTENode.parent` foreign key is updated
            and the `position` callable is called if present), a call to
            :meth:`Model.save` is made.

            :param destination: the destination node of this move, ``None``
                denoting that the node will become a root node.

            :param position: optional callable invoked prior to placement for
                purposes of custom sibling ordering semantics.

            :param save: optional flag indicating whether this model's
                :meth:`save` method should be invoked after the move.

            :return: this node.
        """
        # Allow custom positioning semantics to specify the position before
        # setting the parent.
        if position is not None:
            position(node, destination)
        node.parent = destination
        if save:
            node.save()
        return node


class CTENode(Model):
    """ Abstract :class:`Model` which implements a node in a CTE tree. This
        model features a mandatory foreign key to the parent node (hence to
        ``self``), which, when ``None``, indicates a root node. Multiple nodes
        with a ``None`` parent results in a forest, which can be constrained
        either with custom SQL constraints or through application logic.

        It is necessary for any custom :class:`Manager` of this model to inherit
        from :class:`CTENodeManager`, as all functionality of the CTE tree is
        implemented in the manager.

        It is possible to manipulate individual nodes when not loaded through
        the custom manager, or when freshly created either through the
        :meth:`create` method or through the constructor, however, any operation
        which requires tree information (the :attr:`depth`, :attr:`path`,
        and :attr:`ordering` virtual fields) will not work, and any attempt to
        invoke such methods will result in an :class:`ImproperlyConfigured`
        exception being raised.

        Many runtime properties of nodes are specified through a set of
        parameters which are stored as attributes of the node class, and begin
        with ``_cte_node_``. Before any of these parameters are used, the
        manager will attempt to load and verify them, raising an
        :class:`ImproperlyConfigured` exception if any errors are encountered.
        All parameters have default values.

        All :class:`QuerySet` objects involving CTE nodes use the
        :meth:`QuerySet.extra` semantics in order to specify additional
        ``SELECT``, ``WHERE``, and ``ORDER_BY`` SQL semantics, therefore, they
        cannot be combined through the ``OR`` operator (the ``|`` operator).

        The following parameters can optionally be specified at the class level:

        * _cte_node_traversal:

            A string from one of :const:`TREE_TRAVERSAL_METHODS`, which
            specifies the default tree traversal order. If this parameters is
            ``None`` or :const:`TREE_TRAVERSAL_NONE`, then
            :const:`DEFAULT_TREE_TRAVERSAL` method is used (which is ``dfs``
            for depth-first).

        * _cte_node_order_by:

            A list of strings or tuples specifying the ordering of siblings
            during tree traversal (in the breadth-first method, siblings are
            ordered depending on their parent and not the entire set of nodes at
            the given depth of the tree).

            The entries in this list can be any of the model fields, much like
            the entries in the :attr:`ordering` of the model's :class:`Meta`
            class or the arguments of the :meth:`order_by` method of
            :class:`QuerySet`.

            These entries may also contain the virtual field :attr:`depth`,
            which cannot be used by the normal :class:`QuerySet` because Django
            cannot recognize such virtual fields.

            In case of multiple entries, they must all be of the same database
            type. For VARCHAR fields, their values will be cast to TEXT, unless
            otherwise specified. It is possible to specify the database type
            into which the ordering field values are cast by providing tuples of
            the form ``(fieldname, dbtype)`` in the ordering sequence.

            Specifying cast types is necessary when combining different data
            types in the ordering sequence, such as an int and a float (casting
            the int into a float is probably the desired outcome in this
            situation). In the worst case, TEXT can be specified for all casts.

        * _cte_node_delete_method:

            A string specifying the desired default deletion semantics, which
            may be one of :const:`DELETE_METHODS`. If this parameter is missing
            or ``None`` or :const:`DELETE_METHOD_NONE`, then the default
            deletion semantics :const:`DEFAULT_DELETE_METHOD` will be used
            (which is :const:`DELETE_METHOD_PHARAOH` or ``pharaoh`` for the
            Pharaoh deletion semantics).

        * _cte_node_parent:

            A string referencing the name of the :class:`ForeignKey` field which
            implements the parent relationship, typically called ``parent`` and
            automatically inherited from this class.

            If this parameter is missing, and no field with the name ``parent``
            can be found, then the first :class:`ForeignKey` which relates to
            this model will be used as the parent relationship field.

        * _cte_node_children:

            A string referencing the `related_name` attribute of the
            :class:`ForeignKey` field which implements the parent relationship,
            typically called ``parent`` (specified in
            :const:`DEFAULT_CHILDREN_NAME`) and automatically
            inherited from this class.

        * _cte_node_table:

            The name of the temporary table to use with the ``WITH`` CTE SQL
            statement when compiling queries involving nodes. By default this is
            :const:`DEFAULT_TABLE_NAME` (which is ``cte``).

        * _cte_node_primary_key_type:

            A string representing the database type of the primary key, if the
            primary key is a non-standard type, and must be cast in order to be
            used in the :attr:`path` or :attr:`ordering` virtual fields
            (similarly to the :attr:`_cte_node_order_by` parameter above).

            A ``VARCHAR`` primary key will be automatically cast to ``TEXT``,
            unless explicitly specified otherwise through this parameter.


        * _cte_node_path, _cte_node_depth, _cte_node_ordering:

            Strings specifying the attribute names of the virtual fields
            containing the path, depth, and ordering prefix of each node, by
            default, respectively, :const:`VIRTUAL_FIELD_PATH` (which is
            ``path``), :const:`VIRTUAL_FIELD_DEPTH` (which is ``depth``), and
            :const:`VIRTUAL_FIELD_ORDERING` (which is ``ordering``).
    """

    # This ForeignKey is mandatory, however, its name can be different, as long
    # as it's specified through _cte_node_parent.
    _cte_node_parent = 'parent'
    parent = ForeignKey('self', on_delete=CASCADE, null=True, blank=True,
                        related_name='children')

    # This custom Manager is mandatory.
    objects = CTENodeManager()

    class Meta:
        abstract = True
        base_manager_name = 'objects'
        # Prevent cycles in order to maintain tree / forest property.
        unique_together = [('id', 'parent')]

    def clean(self):
        """ Prevents cycles in the tree. """
        super(CTENode, self).clean()
        if self.parent and self.pk in getattr(self.parent, self._cte_node_path):
            raise ValidationError(_(
                'A node cannot be made a descendant of itself.'
            ))

    def root(self):
        """ Returns the CTENode which is the root of the tree in which this
            node participates.
        """
        return self.__class__.objects.root(self)

    def siblings(self):
        """ Returns a :class:`QuerySet` of all siblings of this node.

            :returns: A :class:`QuerySet` of all siblings of this node.
        """
        return self.__class__.objects.siblings(self)

    def ancestors(self):
        """ Returns a :class:`QuerySet` of all ancestors of this node.

            :returns: A :class:`QuerySet` of all ancestors of this node.
        """
        return self.__class__.objects.ancestors(self)

    def descendants(self):
        """ Returns a :class:`QuerySet` of all descendants of this node.

            :returns: A :class:`QuerySet` of all descendants of this node.
        """
        return self.__class__.objects.descendants(self)

    def is_parent_of(self, subject):
        """ Returns ``True`` if this node is the parent of the given `subject`
            node, ``False`` otherwise. This method uses the :attr:`parent`
            field, and so does not perform any query.

            :param subject: the :class:`CTENode` for which to determine whether
                its parent is this node.

            :returns: ``True`` if this node is the parent of `subject`,
                ``False`` otherwise.
        """
        return self.__class__.objects.is_parent_of(self, subject)

    def is_child_of(self, subject):
        """ Returns ``True`` if this node is a child of the given `subject`
            node, ``False`` otherwise. This method used the :attr:`parent`
            field, and so does not perform any query.

            :param subject: the :class:`CTENode` for which to determine whether
                one of its children is this node.

            :returns: ``True`` if this node is a child of `subject`, ``False``
                otherwise.
        """
        return self.__class__.objects.is_child_of(self, subject)

    def is_sibling_of(self, subject):
        """ Returns ``True`` if this node is a sibling of the given `subject`
            node, ``False`` otherwise. This method uses the :attr:`parent`
            field, and so does not perform any query.

            :param subject: the :class:`CTENode` for which to determine whether
                one of its siblings is this node.

            :returns: ``True`` if this node is a sibling of `subject`, ``False``
                otherwise.
        """
        return self.__class__.objects.is_sibling_of(self, subject)

    def is_ancestor_of(self, subject):
        """ Returns ``True`` if the node is an ancestor of the given `subject`
            node, ``False`` otherwise. This method uses the :attr:`path` virtual
            field, and so does not perform any query.

            :param subject: the :class:`CTENode` for which to determine whether
                one of its ancestors is this node.

            :returns: ``True`` if this node is an ancestor of `subject`,
                ``False`` otherwise.
        """
        return self.__class__.objects.is_ancestor_of(self, subject)

    def is_descendant_of(self, subject):
        """ Returns ``True`` if the node is a descendant of the given `subject`
            node, ``False`` otherwise. This method uses the :attr:`path` virtual
            field, and so does not perform any query.

            :param subject: the :class:`CTENode` for which to determine whether
                one of its descendants is this node.

            :returns: ``True`` if this node is a descendant of `subject`,
                ``False`` otherwise.
        """
        return self.__class__.objects.is_descendant_of(self, subject)

    def is_leaf(self):
        """ Returns ``True`` if this node is a leaf (has no children), ``False``
            otherwise.

            :return: ``True`` if this node is a leaf, ``False`` otherwise.
        """
        return self.__class__.objects.is_leaf(self)

    def is_branch(self):
        """ Returns ``True`` if this node is a branch (has at least one child),
            ``False`` otherwise.

            :return: ``True`` if this node is a branch, ``False`` otherwise.
        """
        return self.__class__.objects.is_branch(self)

    def attribute_path(self, attribute, missing=None, visitor=None):
        """ Generates a list of values of the `attribute` of all ancestors of
            this node (as well as the node itself). If a value is ``None``, then
            the optional value of `missing` is used (by default ``None``).

            By default, the ``getattr(node, attribute, None) or missing``
            mechanism is used to obtain the value of the attribute for each
            node. This can be overridden by supplying a custom `visitor`
            function, which expects as arguments the node and the attribute, and
            should return an appropriate value for the required attribute.

            :param attribute: the name of the attribute.

            :param missing: optional value to use when attribute value is None.

            :param visitor: optional function responsible for obtaining the
                attribute value from a node.

            :return: a list of values of the required `attribute` of the
                ancestor path of this node.
        """
        _parameters = {'node': self, 'attribute': attribute}
        if missing is not None:
            _parameters['missing'] = missing
        if visitor is not None:
            _parameters['visitor'] = visitor
        return self.__class__.objects.attribute_path(**_parameters)

    def as_tree(self, visitor=None, children=None):
        """ Recursively traverses each tree (starting from each root) in order
            to generate a dictionary-based tree structure of the entire forest.
            Each level of the forest/tree is a list of nodes, and each node
            consists of a dictionary representation, where the entry
            ``children`` (by default) consists of a list of dictionary
            representations of its children.

            See :meth:`CTENodeManager.as_tree` and
            :meth:`CTENodeManager.node_as_tree` for details on how this method
            works, as well as its expected arguments.

            :param visitor: optional function responsible for generating the
                dictionary representation of a node.

            :param children: optional function responsible for generating a
                children key and list for a node.

            :return: a dictionary representation of the structure of the forest.
        """
        _parameters = {'node': self}
        if visitor is not None:
            _parameters['visitor'] = visitor
        if children is not None:
            _parameters['children'] = children
        return self.__class__.objects.node_as_tree(**_parameters)

    def move(self, destination=None, position=None, save=False):
        """ Moves this node and places it as a child node of the `destination`
            :class:`CTENode` (or makes it a root node if `destination` is
            ``None``).

            Optionally, `position` can be a callable which is invoked prior to
            placement of the node with this node and the destination node as the
            sole two arguments; this can be useful in implementing specific
            sibling ordering semantics.

            Optionally, if `save` is ``True``, after the move operation
            completes (after the :attr:`parent` foreign key is updated and the
            `position` callable is called if present), a call to
            :meth:`Model.save` is made.

            :param destination: the destination node of this move, ``None``
                denoting that the node will become a root node.

            :param position: optional callable invoked prior to placement for
                purposes of custom sibling ordering semantics.

            :param save: optional flag indicating whether this model's
                :meth:`save` method should be invoked after the move.

            :return: this node.
        """
        return self.__class__.objects.move(self, destination, position, save)

    def delete(self, method=None, position=None, save=True):
        """ Prepares the tree for deletion according to the deletion semantics
            specified for the :class:`CTENode` Model, and then delegates to the
            :class:`CTENode` superclass ``delete`` method.

            Default deletion `method` and `position` callable can be overridden
            by being supplied as arguments to this method.

            :param method: optionally a particular deletion method, overriding
                the default method specified for this model.

            :param position: optional callable to invoke prior to each move
                operation, should the delete method require any moves.

            :param save: optional flag indicating whether this model's
                :meth:`save` method should be invoked after each move operation,
                should the delete method require any moves.
        """
        self.__class__.objects.prepare_delete(self, method=method,
                                              position=position, save=save)
        return super(CTENode, self).delete()



# -*- coding: utf-8 -*-
#
# This document is free and open-source software, subject to the OSI-approved
# BSD license below.
#
# Copyright (c) 2011 - 2013 Alexis Petrounias <www.petrounias.org>,
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the author nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

""" Models for the Django CTE Trees test application.
"""

from uuid import uuid4

from django.db.models import (
    CASCADE, CharField, DateField, FloatField, ForeignKey, Model,
    PositiveIntegerField, UUIDField,
)

from cte_forest.models import CTENode, CTENodeManager


class SimpleNode(CTENode, Model):

    pass


class NoneDeleteNode(CTENode, Model):

    _cte_node_delete_method = 'none'


class SimpleNodeUser(Model):

    node = ForeignKey(SimpleNode, on_delete=CASCADE, null=False)


class NamedNode(CTENode, Model):

    name = CharField(max_length=128, null=False)

    class Meta(object):

        abstract = True


class SimpleNamedNode(NamedNode):

    pass


class OrderedNamedNode(NamedNode):

    _cte_node_order_by = ['name']


class ValueNamedNode(NamedNode):

    v = PositiveIntegerField()


class SimpleNamedNodeUser(Model):

    node = ForeignKey(SimpleNamedNode, on_delete=CASCADE, null=False)

    name = CharField(max_length=128, null=False)


class DFSOrderedNode(CTENode, Model):

    v = PositiveIntegerField()

    _cte_node_traversal = 'dfs'

    _cte_node_order_by = ['v']


class BFSOrderedNode(CTENode, Model):

    v = PositiveIntegerField()

    _cte_node_traversal = 'bfs'

    _cte_node_order_by = ['v']


class NoneTraversalNode(CTENode, Model):

    v = PositiveIntegerField()

    _cte_node_traversal = 'none'

    _cte_node_order_by = ['v']


class TypeCoercionNode(CTENode, Model):

    name = CharField(max_length=128, null=False)

    v = PositiveIntegerField()

    _cte_node_order_by = [('v', 'text'), 'name']


class TypeCombinationNode(CTENode, Model):

    v1 = PositiveIntegerField()

    v2 = FloatField()

    _cte_node_traversal = 'bfs'

    _cte_node_order_by = [('v1', 'float'), 'v2']


class ExoticTypeNode(CTENode, Model):

    v = DateField()

    y = DateField(null=True)

    _cte_node_order_by = ['v']


class DBTypeNode(CTENode, Model):

    v = UUIDField(default=uuid4)

    _cte_node_order_by = ['v']


class CustomPrimaryKeyNode(CTENode, Model):

    id = CharField(max_length=128, primary_key=True)


class DBTypePrimaryKeyNode(CTENode, Model):

    id = UUIDField(primary_key=True, default=uuid4)


class AggregationNode(CTENode, Model):

    price = PositiveIntegerField()


class BadParameter_parent_1_Node(CTENode, Model):

    _cte_node_parent = 'wrong'


class ArbitraryNode(CTENode, Model):

    pass


class BadParameter_parent_2_Node(CTENode, Model):

    wrong = ForeignKey(ArbitraryNode, on_delete=CASCADE, null=True)

    _cte_node_parent = 'wrong'


class ArbitraryModel(Model):

    pass


class BadParameter_parent_3_Node(CTENode, Model):

    wrong = ForeignKey(ArbitraryModel, on_delete=CASCADE, null=True)

    _cte_node_parent = 'wrong'


class BadParameter_parent_4_Node(Model):

    objects = CTENodeManager()


class BadParameter_traversal_Node(CTENode, Model):

    _cte_node_traversal = 'wrong'


class BadParameter_delete_Node(CTENode, Model):

    _cte_node_delete_method = 'wrong'



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This document is free and open-source software, subject to the OSI-approved
# BSD license below.
#
# Copyright (c) 2011 - 2013 Alexis Petrounias <www.petrounias.org>,
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the author nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


import os
from io import open

from setuptools import find_packages, setup


def read(filename):
    path = os.path.join(os.path.dirname(__file__), filename)
    with open(path, encoding='utf-8') as handle:
        return handle.read()


setup(
    name='django-cte-forest',
    version=__import__('cte_forest').__version__,
    description=(
        'Django Adjacency-List trees using PostgreSQL'
        ' Common Table Expressions (CTE).'
    ),
    long_description=read('README.rst'),
    maintainer='Matthias Kestenholz',
    maintainer_email='mk@feinheit.ch',
    url='https://github.com/matthiask/django-cte-forest',
    license='BSD License',
    packages=find_packages(
        exclude=['cte_forest_test'],
    ),
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    zip_safe=False,
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This document is free and open-source software, subject to the OSI-approved
# BSD license below.
#
# Copyright (c) 2011 - 2013 Alexis Petrounias <www.petrounias.org>,
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the author nor the names of its contributors may be used
# to endorse or promote products derived from this software without specific
# prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

""" Dummy Django application settings so unittest and Sphinx autodoc can work.
"""

# Python
import os
import sys

if __name__ == "__main__":
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "cte_forest_test.settings")
    from django.core.management import execute_from_command_line
    execute_from_command_line(sys.argv)
